import re
from collections import defaultdict

idds = ('00', '+')
phone_reg = re.compile(r"(\+|00)?\d[\d\-()]{7,12}\d$")
phone_dial = re.compile("[+(]?(\d{0,6})")  # any prefix like `+420` or `(089)`, then up to 6 digits (there is no longer prefix)


def phone_regex_match(val):
    val = val.replace(" ", "")
    return bool(phone_reg.match(val))


def phone_country(val, get_prefix=False):
    """
    We are doing our best to identify any phone number country.
    Possible formats:
        420 (country code only)
        +420 (plus sign)
        00420 (IDD prefix)
    (Library phonenumbers did not find ex Puerto Rico at "1787")
    XX get_prefix seems to not be used
    """
    for idd in idds:
        if val.startswith(idd):
            val = val.lstrip(idd)
            break
    m = phone_dial.match(val.strip())  # strip the number to what may be its prefix only
    if m:
        val = m.group(1)

    for i in range(1, len(val) + 1):
        # get the biggest possible prefix,
        # ex: 4 → "42 exists" → skipped
        #     42 → "420 exist" → skipped
        #     420 → "420X does not exist" → taken
        # ex: number +1675, note there is +1 for US/CA and +1670 for an island
        #     1 , skipped, 16 skipped, 167 skipped, 1675 taken
        if i < len(val) and int(val[:i]) in phones_prefixes:
            continue
        for i2 in range(i, 0, -1):
            # ex: 1675 does not exist, roll back to 167, 16, 1 which is US/CA
            prefix = int(val[:i2])
            if prefix not in phones:  # this prefix does not exist, must be shorter
                continue

            if get_prefix:
                return prefix
            return phones[prefix]


def address_country_lowered(val):
    for s in countries:
        if s in val:
            return countries[s]


def address_country(val):
    val = val.lower()
    for s in countries:
        if s in val:
            return countries[s]


phones = {1: ['ca', 'us'], 1340: 'us', 1670: 'Northern Mariana Islands', 1671: 'gu', 1684: 'as', 1787: 'pr', 1939: 'pr', 1242: 'bs',
          1246: 'bb', 1264: 'ai', 1268: 'ag', 1284: 'vg', 1345: 'ky', 1441: 'bm', 1473: 'gd', 1649: 'tc', 1664: 'ms', 1721: 'sx',
          1758: 'lc', 1767: 'dm', 1784: 'vc', 1809: 'dm', 1829: 'dm', 1849: 'dm', 1868: 'tt', 1869: 'kn', 1876: 'jm', 1658: 'jm',
          20: 'eg', 211: 'sd', 212: 'ma', 213: 'dz', 216: 'tn', 218: 'ly', 220: 'gm', 221: 'sn', 222: 'mr', 223: 'ml', 224: 'gn',
          225: 'ci', 226: 'bf', 227: 'ne', 228: 'tg', 229: 'bj', 230: 'mu', 231: 'lr', 232: 'sl', 233: 'gh', 234: 'ne', 235: 'td',
          236: 'cf', 237: 'cm', 238: 'cv', 239: 'st', 240: 'gn', 241: 'ga', 242: 'cg', 243: 'cd', 244: 'ao', 245: 'gn', 246: 'in',
          247: 'sh', 248: 'sc', 249: 'sd', 250: 'rw', 251: 'et', 252: 'ml', 253: 'dj', 254: 'ke', 255: 'tz',
          524: 'Zanzibar, in place of never-implemented +259', 256: 'ug', 257: 'bi', 258: 'mz', 260: 'zm', 261: 'mg', 262: 're',
          2269: 'yt', 2639: 'km', 263: 'zw', 264: 'na', 265: 'mw', 266: 'ls', 267: 'bw', 268: 'sz', 269: 'km', 27: 'za', 290: 'sh',
          2908: 'sh', 291: 'er', 297: 'aw', 298: 'fo', 299: 'gl', 30: 'gr', 31: 'nl', 32: 'be', 33: 'fr', 34: 'es', 350: 'gi',
          351: 'pt', 352: 'lu', 353: 'ie', 354: 'is', 355: 'al', 356: 'mt', 357: 'cy', 358: 'fi', 818: 'Åland Islands', 359: 'bg',
          36: 'hu', 370: 'lt', 371: 'lv', 372: 'ee', 373: 'md', 374: 'am', 447: 'Artsakh (landlines, formerly +7 893)',
          497: 'Artsakh (mobile phones)', 375: 'by', 376: 'ad', 377: 'mc', 378: 'sm', 379: 'it', 380: 'ua', 381: 'rs', 382: 'me',
          383: 'Kosovo', 385: 'hr', 386: 'si', 387: 'ba', 389: 'mk', 39: 'it',
          906698: ' Vatican City (assigned +379 but not in use)', 40: 'om', 41: 'ch', 420: 'cz', 421: 'sk', 423: 'li', 43: 'at',
          44: 'gb', 41481: 'gg', 41534: 'je', 41624: 'im', 45: 'dk', 46: 'se', 47: 'no', 779: 'sj', 48: 'pl', 49: 'de', 500: 'fk',
          501: 'bz', 502: 'gt', 503: 'sv', 504: 'hn', 505: 'ni', 506: 'cr', 507: 'pa', 508: 'pm', 509: 'ht', 51: 'pe', 52: 'mx',
          53: 'cu', 54: 'ar', 55: 'br', 56: 'cl', 57: 'co', 58: 've', 590: 'gp', 591: 'bo', 592: 'gy', 593: 'ec', 594: 'fr',
          595: 'py', 596: 'mq', 597: 'sr', 598: 'uy', 5993: 'bq', 5994: 'bq', 5997: 'bq', 5999: 'cw', 60: 'my', 61: 'au',
          189162: 'cc', 189164: 'cx', 62: 'id', 63: 'ph', 64: 'pn', 65: 'sg', 66: 'th', 670: 'id', 23: 'nf', 673: 'bn', 674: 'nr',
          675: 'gn', 676: 'to', 677: 'sb', 678: 'vu', 679: 'fj', 680: 'pw', 681: 'wf', 682: 'ck', 683: 'nu', 685: 'ws', 686: 'ki',
          687: 'nc', 688: 'tv', 689: 'fr', 690: 'tk', 691: 'fm', 692: 'mh', 7: 'ru', 76: 'kz', 77: 'kz',
          7840: 'Abkhazia - see also +995 44', 7940: 'Abkhazia - see also +995 44', 800: 'International Freephone (UIFN)',
          808: 'reserved for Shared Cost Services', 81: 'jp', 82: 'kr', 84: 'vn', 850: 'kp', 852: 'hk', 853: 'mo', 855: 'kh',
          856: 'la', 86: 'cn', 870: 'Inmarsat "SNAC" service', 875: 'Maritime Mobile service', 876: 'Maritime Mobile service',
          877: 'Maritime Mobile service', 878: 'Universal Personal Telecommunications services',
          879: 'national non-commercial purposes', 880: 'bd', 881: 'Global Mobile Satellite System', 882: 'International Networks',
          883: 'International Networks', 886: 'tw', 888: 'Telecommunications for Disaster Relief by OCHA', 90: 'tr', 90392: 'cy',
          91: 'in', 92: 'pk', 2582: 'Azad Kashmir[3]', 2581: 'Gilgit Baltistan[3]', 93: 'af', 94: 'lk', 95: 'Myanmar', 960: 'mv',
          961: 'lb', 962: 'jo', 963: 'sy', 964: 'iq', 965: 'kw', 966: 'sa', 967: 'ye', 968: 'om', 970: 'ps', 971: 'ae', 972: 'il',
          973: 'bh', 974: 'qa', 975: 'bt', 976: 'mn', 977: 'np',
          979: 'International Premium Rate Service - originally assigned to Abu Dhabi, now covered under 971', 98: 'ir',
          991: 'International Telecommunications Public Correspondence Service trial (ITPCS)', 992: 'tj', 993: 'tm', 994: 'az',
          995: 'ge', 534: 'South Ossetia', 544: 'Abkhazia[4][5] - see also +7 840, 940', 996: 'kg', 998: 'uz'}

phones_prefixes = defaultdict(list)
for k in phones:
    k = str(k)
    for i in range(1, len(k)):
        try:
            phones_prefixes[int(k[:i])].append(int(k[i]))
        except IndexError:
            pass

countries = {"andorra": "ad",
             "united arab emirates": "ae",
             "afghanistan": "af",
             "barbuda": "ag",
             "antigua": "ag",
             "anguilla": "ai",
             "albania": "al",
             "armenia": "am",
             "angola": "ao",
             "antarctica ": "aq",
             "argentine": "ar",
             "argentina": "ar",
             "american samoa": "as",
             "austria": "at",
             "australia": "au",
             "australia ": "au",
             "aruba": "aw",
             "aland": "ax",
             "azerbaijan": "az",
             "herzegovina": "ba",
             "bosnia": "ba",
             "barbados": "bb",
             "bangladesh": "bd",
             "belgium": "be",
             "burkina faso": "bf",
             "bulgaria": "bg",
             "bahrain": "bh",
             "burundi": "bi",
             "benin": "bj",
             "barthelemy": "bl",
             "bermuda": "bm",
             "brunei": "bn",
             "bolivia": "bo",
             "bonaire": "bq",
             "eustatius": "bq",
             "saba": "bq",
             "brazil": "br",
             "bahamas": "bs",
             "bhutan": "bt",
             "bouvet": "bv",
             "botswana": "bw",
             "belarus": "by",
             "belize": "bz",
             "canada": "ca",
             "keeling": "cc",
             "cocos": "cc",
             "democratic republic of the congo": "cd",
             "central african": "cf",
             "congo": "cg",
             "swiss": "ch",
             "switzerland": "ch",
             "côte d'ivoire": "ci",
             "ivory coast": "ci",
             "cook": "ck",
             "chile": "cl",
             "cameroon": "cm",
             "china": "cn",
             "colombia": "co",
             "costa rica": "cr",
             "cuba": "cu",
             "cabo verde": "cv",
             "cabo verde ": "cv",
             "cape verde": "cv",
             "curacao": "cw",
             "curaçao": "cw",
             "christmas island": "cx",
             "cyprus": "cy",
             "czech": "cz",
             "germany": "de",
             "djibouti": "dj",
             "denmark": "dk",
             "dominica": "dm",
             "dominican": "do",
             "algeria": "dz",
             "ecuador": "ec",
             "estonia": "ee",
             "egypt": "eg",
             "sahrawi": "eh",
             "western sahara": "eh",
             "eritrea": "er",
             "spain": "es",
             "ethiopia": "et",
             "finland": "fi",
             "fiji": "fj",
             "falkland": "fk",
             "falkland islands": "fk",
             "micronesia": "fm",
             "faroe": "fo",
             "french": "fr",
             "france": "fr",
             "gabonese": "ga",
             "gabon": "ga",
             "northern ireland": "gb",
             "great britain": "gb",
             "united kingdom": "gb",
             "england": "gb",
             "grenada": "gd",
             "georgia": "ge",
             "guyane": "gf",
             "french guiana": "gf",
             "bailiwick of guernsey": "gg",
             "guernsey": "gg",
             "ghana": "gh",
             "gibraltar": "gi",
             "kalaallit nunaat": "gl",
             "greenland": "gl",
             "gambia": "gm",
             "guinea": "gn",
             "guadeloupe": "gp",
             "equatorial guinea": "gq",
             "hellenic": "gr",
             "greece": "gr",
             "south sandwich": "gs",
             "south georgia": "gs",
             "guatemala": "gt",
             "guam": "gu",
             "guinea-bissau": "gw",
             "guyana": "gy",
             "hong kong": "hk",
             "mcdonald": "hm",
             "heard": "hm",
             "honduras": "hn",
             "croatia": "hr",
             "haiti": "ht",
             "hungary": "hu",
             "indonesia": "id",
             "ireland": "ie",
             "israel": "il",
             "isle of man": "im",
             "india": "in",
             "british indian ocean": "io",
             "iraq": "iq",
             "iran": "ir",
             "iceland": "is",
             "italian": "it",
             "italy": "it",
             "jersey": "je",
             "jamaica": "jm",
             "hashemite": "jo",
             "jordan": "jo",
             "japan": "jp",
             "kenya": "ke",
             "kyrgyz": "kg",
             "cambodia": "kh",
             "kiribati": "ki",
             "comoros": "km",
             "nevis": "kn",
             "saint kitts": "kn",
             "democratic people": "kp",
             "north korea": "kp",
             "korea": "kr",
             "kuwait": "kw",
             "cayman": "ky",
             "kazakhstan": "kz",
             "lao": "la",
             "lebanese": "lb",
             "lebanon": "lb",
             "lucia": "lc",
             "saint lucia": "lc",
             "liechtenstein": "li",
             "sri lanka": "lk",
             "liberia": "lr",
             "lesotho": "ls",
             "lithuania": "lt",
             "luxembourg": "lu",
             "latvia": "lv",
             "libya": "ly",
             "morocco": "ma",
             "monaco": "mc",
             "moldova": "md",
             "montenegro": "me",
             "saint-martin": "mf",
             "saint martin": "mf",
             "madagascar": "mg",
             "marshall islands": "mh",
             "north macedonia": "mk",
             "mali": "ml",
             "myanmar ": "mm",
             "mongolia": "mn",
             "macao": "mo",
             "macau": "mo",
             "norrn mariana islands": "mp",
             "martinique": "mq",
             "mauritania": "mr",
             "montserrat": "ms",
             "malta": "mt",
             "mauritius": "mu",
             "maldives": "mv",
             "malawi": "mw",
             "mexican": "mx",
             "mexico": "mx",
             "malaysia": "my",
             "mozambique": "mz",
             "namibia": "na",
             "new caledonia": "nc",
             "niger": "ne",
             "norfolk island": "nf",
             "federal nigeria": "ng",
             "nigeria": "ng",
             "nicaragua": "ni",
             "netherlands": "nl",
             "norway": "no",
             "nepal": "np",
             "nauru": "nr",
             "niue": "nu",
             "new zealand": "nz",
             "oman": "om",
             "panama": "pa",
             "peru": "pe",
             "french polynesia": "pf",
             "new guinea": "pg",
             "papua": "pg",
             "philippines": "ph",
             "pakistan": "pk",
             "poland": "pl",
             "miquelon": "pm",
             "saint-pierre": "pm",
             "pitcairn": "pn",
             "henderson": "pn",
             "ducie": "pn",
             "oeno": "pn",
             "puerto rico": "pr",
             "palestine": "ps",
             "portuguese": "pt",
             "portugal": "pt",
             "palau": "pw",
             "paraguay": "py",
             "qatar": "qa",
             "reunion": "re",
             "réunion": "re",
             "romania": "ro",
             "serbia": "rs",
             "russia": "ru",
             "rwanda": "rw",
             "saudi arabia": "sa",
             "solomon islands": "sb",
             "seychelles": "sc",
             "sudan": "sd",
             "sweden": "se",
             "singapore": "sg",
             "helena": "sh",
             "ascension": "sh",
             "tristan": "sh",
             "cunha": "sh",
             "slovenia": "si",
             "jan mayen": "sj",
             "svalbard": "sj",
             "slovak": "sk",
             "slovakia": "sk",
             "sierra leone": "sl",
             "san marino": "sm",
             "senegal": "sn",
             "federal somalia": "so",
             "somalia": "so",
             "suriname": "sr",
             "south sudan": "ss",
             "príncipe": "st",
             "sao tome": "st",
             "principe": "st",
             "el salvador": "sv",
             "sint maarten": "sx",
             "syria": "sy",
             "eswatini": "sz",
             "eswatini ": "sz",
             "caicos islands": "tc",
             "turks": "tc",
             "chad": "td",
             "antarctic lands": "tf",
             "french southern": "tf",
             "togolese": "tg",
             "togo": "tg",
             "thailand": "th",
             "tajikistan": "tj",
             "tokelau": "tk",
             "timor-leste": "tl",
             "turkmenistan": "tm",
             "tunisia": "tn",
             "tonga": "to",
             "turkey": "tr",
             "tobago": "tt",
             "trinidad": "tt",
             "tuvalu": "tv",
             "taiwan": "tw",
             "tanzania": "tz",
             "ukraine": "ua",
             "uganda": "ug",
             "baker": "um",
             "howland": "um",
             "jarvis": "um",
             "johnston": "um",
             "kingman": "um",
             "midway": "um",
             "navassa": "um",
             "palmyra": "um",
             "wake": "um",
             "minor outlying islands  ": "um",
             "united states": "us",
             "uruguay": "uy",
             "uzbekistan": "uz",
             "holy see": "va",
             "grenadines": "vc",
             "saint vincent": "vc",
             "venezuela": "ve",
             "british virgin islands": "vg",
             "virgin islands of united states": "vi",
             "viet nam ": "vn",
             "vietnam": "vn",
             "vanuatu": "vu",
             "futuna": "wf",
             "wallis": "wf",
             "samoa": "ws",
             "yemen": "ye",
             "mayotte": "yt",
             "south africa": "za",
             "zambia": "zm",
             "zimbabwe": "zw"}

country_codes = defaultdict(str, {'af': 'Afghanistan',
                                  'ax': 'Åland Islands',
                                  'al': 'Albania',
                                  'dz': 'Algeria',
                                  'as': 'American Samoa',
                                  'ad': 'Andorra',
                                  'ao': 'Angola',
                                  'ai': 'Anguilla',
                                  'aq': 'Antarctica',
                                  'ag': 'Antigua and Barbuda',
                                  'ar': 'Argentina',
                                  'am': 'Armenia',
                                  'aw': 'Aruba',
                                  'au': 'Australia',
                                  'at': 'Austria',
                                  'az': 'Azerbaijan',
                                  'bs': 'Bahamas',
                                  'bh': 'Bahrain',
                                  'bd': 'Bangladesh',
                                  'bb': 'Barbados',
                                  'by': 'Belarus',
                                  'be': 'Belgium',
                                  'bz': 'Belize',
                                  'bj': 'Benin',
                                  'bm': 'Bermuda',
                                  'bt': 'Bhutan',
                                  'bo': 'Bolivia (Plurinational State of)',
                                  'bq': 'Bonaire, Sint Eustatius and Saba',
                                  'ba': 'Bosnia and Herzegovina',
                                  'bw': 'Botswana',
                                  'bv': 'Bouvet Island',
                                  'br': 'Brazil',
                                  'io': 'British Indian Ocean Territory',
                                  'bn': 'Brunei Darussalam',
                                  'bg': 'Bulgaria',
                                  'bf': 'Burkina Faso',
                                  'bi': 'Burundi',
                                  'cv': 'Cabo Verde',
                                  'kh': 'Cambodia',
                                  'cm': 'Cameroon',
                                  'ca': 'Canada',
                                  'ky': 'Cayman Islands',
                                  'cf': 'Central African Republic',
                                  'td': 'Chad',
                                  'cl': 'Chile',
                                  'cn': 'China',
                                  'cx': 'Christmas Island',
                                  'cc': 'Cocos (Keeling) Islands',
                                  'co': 'Colombia',
                                  'km': 'Comoros',
                                  'cg': 'Congo',
                                  'cd': 'Congo, Democratic Republic of the',
                                  'ck': 'Cook Islands',
                                  'cr': 'Costa Rica',
                                  'ci': "Côte d'Ivoire",
                                  'hr': 'Croatia',
                                  'cu': 'Cuba',
                                  'cw': 'Curaçao',
                                  'cy': 'Cyprus',
                                  'cz': 'Czechia',
                                  'dk': 'Denmark',
                                  'dj': 'Djibouti',
                                  'dm': 'Dominica',
                                  'do': 'Dominican Republic',
                                  'ec': 'Ecuador',
                                  'eg': 'Egypt',
                                  'sv': 'El Salvador',
                                  'gq': 'Equatorial Guinea',
                                  'er': 'Eritrea',
                                  'ee': 'Estonia',
                                  'sz': 'Eswatini',
                                  'et': 'Ethiopia',
                                  'fk': 'Falkland Islands (Malvinas)',
                                  'fo': 'Faroe Islands',
                                  'fj': 'Fiji',
                                  'fi': 'Finland',
                                  'fr': 'France',
                                  'gf': 'French Guiana',
                                  'pf': 'French Polynesia',
                                  'tf': 'French Southern Territories',
                                  'ga': 'Gabon',
                                  'gm': 'Gambia',
                                  'ge': 'Georgia',
                                  'de': 'Germany',
                                  'gh': 'Ghana',
                                  'gi': 'Gibraltar',
                                  'gr': 'Greece',
                                  'gl': 'Greenland',
                                  'gd': 'Grenada',
                                  'gp': 'Guadeloupe',
                                  'gu': 'Guam',
                                  'gt': 'Guatemala',
                                  'gg': 'Guernsey',
                                  'gn': 'Guinea',
                                  'gw': 'Guinea-Bissau',
                                  'gy': 'Guyana',
                                  'ht': 'Haiti',
                                  'hm': 'Heard Island and McDonald Islands',
                                  'va': 'Holy See',
                                  'hn': 'Honduras',
                                  'hk': 'Hong Kong',
                                  'hu': 'Hungary',
                                  'is': 'Iceland',
                                  'in': 'India',
                                  'id': 'Indonesia',
                                  'ir': 'Iran (Islamic Republic of)',
                                  'iq': 'Iraq',
                                  'ie': 'Ireland',
                                  'im': 'Isle of Man',
                                  'il': 'Israel',
                                  'it': 'Italy',
                                  'jm': 'Jamaica',
                                  'jp': 'Japan',
                                  'je': 'Jersey',
                                  'jo': 'Jordan',
                                  'kz': 'Kazakhstan',
                                  'ke': 'Kenya',
                                  'ki': 'Kiribati',
                                  'kp': "Korea (Democratic People's Republic of)",
                                  'kr': 'Korea, Republic of',
                                  'kw': 'Kuwait',
                                  'kg': 'Kyrgyzstan',
                                  'la': "Lao People's Democratic Republic",
                                  'lv': 'Latvia',
                                  'lb': 'Lebanon',
                                  'ls': 'Lesotho',
                                  'lr': 'Liberia',
                                  'ly': 'Libya',
                                  'li': 'Liechtenstein',
                                  'lt': 'Lithuania',
                                  'lu': 'Luxembourg',
                                  'mo': 'Macao',
                                  'mg': 'Madagascar',
                                  'mw': 'Malawi',
                                  'my': 'Malaysia',
                                  'mv': 'Maldives',
                                  'ml': 'Mali',
                                  'mt': 'Malta',
                                  'mh': 'Marshall Islands',
                                  'mq': 'Martinique',
                                  'mr': 'Mauritania',
                                  'mu': 'Mauritius',
                                  'yt': 'Mayotte',
                                  'mx': 'Mexico',
                                  'fm': 'Micronesia (Federated States of)',
                                  'md': 'Moldova, Republic of',
                                  'mc': 'Monaco',
                                  'mn': 'Mongolia',
                                  'me': 'Montenegro',
                                  'ms': 'Montserrat',
                                  'ma': 'Morocco',
                                  'mz': 'Mozambique',
                                  'mm': 'Myanmar',
                                  'na': 'Namibia',
                                  'nr': 'Nauru',
                                  'np': 'Nepal',
                                  'nl': 'Netherlands',
                                  'nc': 'New Caledonia',
                                  'nz': 'New Zealand',
                                  'ni': 'Nicaragua',
                                  'ne': 'Niger',
                                  'ng': 'Nigeria',
                                  'nu': 'Niue',
                                  'nf': 'Norfolk Island',
                                  'mk': 'North Macedonia',
                                  'mp': 'Northern Mariana Islands',
                                  'no': 'Norway',
                                  'om': 'Oman',
                                  'pk': 'Pakistan',
                                  'pw': 'Palau',
                                  'ps': 'Palestine, State of',
                                  'pa': 'Panama',
                                  'pg': 'Papua New Guinea',
                                  'py': 'Paraguay',
                                  'pe': 'Peru',
                                  'ph': 'Philippines',
                                  'pn': 'Pitcairn',
                                  'pl': 'Poland',
                                  'pt': 'Portugal',
                                  'pr': 'Puerto Rico',
                                  'qa': 'Qatar',
                                  're': 'Réunion',
                                  'ro': 'Romania',
                                  'ru': 'Russian Federation',
                                  'rw': 'Rwanda',
                                  'bl': 'Saint Barthélemy',
                                  'sh': 'Saint Helena, Ascension and Tristan da Cunha',
                                  'kn': 'Saint Kitts and Nevis',
                                  'lc': 'Saint Lucia',
                                  'mf': 'Saint Martin (French part)',
                                  'pm': 'Saint Pierre and Miquelon',
                                  'vc': 'Saint Vincent and the Grenadines',
                                  'ws': 'Samoa',
                                  'sm': 'San Marino',
                                  'st': 'Sao Tome and Principe',
                                  'sa': 'Saudi Arabia',
                                  'sn': 'Senegal',
                                  'rs': 'Serbia',
                                  'sc': 'Seychelles',
                                  'sl': 'Sierra Leone',
                                  'sg': 'Singapore',
                                  'sx': 'Sint Maarten (Dutch part)',
                                  'sk': 'Slovakia',
                                  'si': 'Slovenia',
                                  'sb': 'Solomon Islands',
                                  'so': 'Somalia',
                                  'za': 'South Africa',
                                  'gs': 'South Georgia and the South Sandwich Islands',
                                  'ss': 'South Sudan',
                                  'es': 'Spain',
                                  'lk': 'Sri Lanka',
                                  'sd': 'Sudan',
                                  'sr': 'Suriname',
                                  'sj': 'Svalbard and Jan Mayen',
                                  'se': 'Sweden',
                                  'ch': 'Switzerland',
                                  'sy': 'Syrian Arab Republic',
                                  'tw': 'Taiwan, Province of China[a]',
                                  'tj': 'Tajikistan',
                                  'tz': 'Tanzania, United Republic of',
                                  'th': 'Thailand',
                                  'tl': 'Timor-Leste',
                                  'tg': 'Togo',
                                  'tk': 'Tokelau',
                                  'to': 'Tonga',
                                  'tt': 'Trinidad and Tobago',
                                  'tn': 'Tunisia',
                                  'tr': 'Turkey',
                                  'tm': 'Turkmenistan',
                                  'tc': 'Turks and Caicos Islands',
                                  'tv': 'Tuvalu',
                                  'ug': 'Uganda',
                                  'ua': 'Ukraine',
                                  'ae': 'United Arab Emirates',
                                  'gb': 'United Kingdom of Great Britain and Northern Ireland',
                                  'us': 'United States of America',
                                  'um': 'United States Minor Outlying Islands',
                                  'uy': 'Uruguay',
                                  'uz': 'Uzbekistan',
                                  'vu': 'Vanuatu',
                                  've': 'Venezuela (Bolivarian Republic of)',
                                  'vn': 'Viet Nam',
                                  'vg': 'Virgin Islands (British)',
                                  'vi': 'Virgin Islands (U.S.)',
                                  'wf': 'Wallis and Futuna',
                                  'eh': 'Western Sahara',
                                  'ye': 'Yemen',
                                  'zm': 'Zambia',
                                  'zw': 'Zimbabwe'})



import logging
import re
from collections import OrderedDict
from datetime import datetime, timedelta
from subprocess import PIPE, Popen
from time import time, sleep

from netaddr import IPRange, IPNetwork

from .contacts import Contacts
from .config import Config, subprocess_env
from .infodicts import address_country_lowered

logger = logging.getLogger(__name__)

rirs = ["whois.ripe.netf", "whois.arin.net", "whois.lacnic.net", "whois.apnic.net", "whois.afrinic.net"]
whole_space = IPRange('0.0.0.0', '255.255.255.255')


class Quota:
    def __init__(self):
        self._time = None

    def try_start(self):
        if not self._time:
            self._time = datetime.now() + timedelta(seconds=300)

    def time(self):
        return self._time.strftime('%H:%M')

    def is_running(self):
        return self._time and self._time > datetime.now()

    def check_over(self):
        if self._time and self._time < datetime.now():
            self._time = None
            Whois.queued_ips = set()

    def remains(self):
        if self.is_running():
            return (self._time - datetime.now()).seconds

    class QuotaExceeded(IOError):
        pass


class UnknownValue(LookupError):
    pass


class Whois:
    slow_mode: bool
    unknown_mode: bool
    quota: Quota
    queued_ips: set
    see: int

    @classmethod
    def init(cls, stats, ranges, ip_seen, csvstats, slow_mode=False, unknown_mode=False):
        cls.quota = Quota()
        cls.csvstats = csvstats
        cls.stats = stats
        cls.ranges = ranges
        cls.ip_seen = ip_seen  # ip_seen[ip] = prefix
        cls.servers = OrderedDict()
        cls.unknown_mode = unknown_mode  # if True, we use b flag in abusemails
        cls.slow_mode = slow_mode  # due to LACNIC quota
        cls.queued_ips = set()
        cls.ttl = Config.get("whois_ttl", "FIELDS", int)
        cls.see = Config.verbosity <= logging.INFO
        if Config.get("whois_mirror", "FIELDS"):  # try a fast local whois-mirror first
            cls.servers["mirror"] = Config.get("whois_mirror", "FIELDS")
        cls.servers["general"] = None
        # Algorithm for querying custom servers:
        # for name, val in zip(["ripe", "arin", "lacnic", "apnic", "afrinic"],
        #                      ["whois.ripe.net -r", "whois.arin.net", "whois.lacnic.net", "whois.apnic.net", "whois.afrinic.net"]):
        #     Whois.servers[name] = val

    def __init__(self, ip):
        """
         self.get stores tuple: prefix, location, mail, asn, netname, country, ttl
        """
        self.ip = ip
        self.whois_response = []
        prefix = self.cache_load()  # try load prefix from earlier WHOIS responses
        if prefix:
            if (self.ttl != -1 and self.get[7] + self.ttl < time()) or (Whois.unknown_mode and not self.get[6]):
                # the TTL is too old, we cannot guarantee IP stayed in the same prefix, let's get rid of the old results
                # OR we are in unknown_mode which means we want abusemail. If not here, maybe another IP claimed
                # a range superset without abuse e-mail. Delete this possible superset
                # We do not have to call now `self.get = None; del self.ip_seen[ip]` if there is no need to be thread safe,
                #   these lines will be called at the function end.
                del self.ranges[prefix]
            else:
                self.count_stats()
                return

        if self.see:
            print(f"Whois {ip}... ", end="", flush=True)
        if Whois.slow_mode:
            if self.see:
                print("waiting 7 seconds... ", end="", flush=True)
            sleep(7)
        get = self.analyze()  # prefix, location, mail, asn, netname, country...
        if self.see:
            print(get[2] or "no incident contact.")
        prefix = get[0]
        if not prefix:
            logger.info(f"No prefix found for IP {ip}")
            prefix = IPRange(0, 0)  # make key consistent when saving into cache
        self.ip_seen[ip] = prefix
        self.get = self.ranges[prefix] = get
        self.count_stats()

    def cache_load(self):
        if self.ip in self.ip_seen:  # ip has been seen in the past
            prefix = self.ip_seen[self.ip]
            if prefix not in self.ranges:  # removed ex: due to TTL
                return
            self.get = self.ranges[prefix]
            return prefix
        elif self.ip in self.queued_ips:
            raise self.quota.QuotaExceeded
        for prefix in self.ranges:
            # search for prefix the slow way. I dont know how to make this shorter because IP can be in shortened form so that
            # in every case I had to put it in full form and then slowly compare strings with prefixes.
            if prefix and self.ip in prefix:
                self.get = self.ranges[prefix]
                self.ip_seen[self.ip] = prefix
                return prefix

    def count_stats(self):
        self.csvstats["ip_unique"].add(self.ip)
        mail = self.get[6]
        contact = self.get[2]
        reg = self.get[1]
        known = "known" if mail else "unknown"
        self.csvstats[f"ip_{reg}_{known}"].add(self.ip)
        self.csvstats[f"prefix_{reg}_{known}"].add(self.get[0])

        if mail:
            self.csvstats[f"abusemail_{reg}"].add(mail)

        if reg == "abroad":
            country = self.get[5]
            if country in Contacts.country2mail:
                known = "known"
            elif mail:
                known = "unofficial"
                self.csvstats[f"abusemail_{known}"].add(mail)  # subset of abusemail_abroad
                self.csvstats[f"prefix_csirtmail_{known}"].add(self.get[0])  # subset of prefix_abroad_un/known
            else:  # we do not track the amount of unknown IP addresses that should be delivered to countries
                known = None

            if known:
                self.csvstats[f"ip_csirtmail_{known}"].add(self.ip)
                self.csvstats[f"csirtmail_{known}"].add(country)

        if not mail and Config.get("whois_reprocessable_unknown", "FIELDS", get=bool):
            raise UnknownValue

    def resolve_unknown_mail(self):
        """ Forces to load abusemail for an IP.
        We try first omit -r flag and then add -B flag.

        XX -B flag disabled (at least temporarily). Document.

        XX Note that we try only RIPE server because it's the only one that has flags -r and -B.
        If ARIN abusemail is not found, we have no help yet. I dont know if that ever happens.
            XX We prefer general calling of whois program instead of asking him for different whois servers manually
            so I'm not sure if whois program calls RIPE with -r by default or not.
            If not, we should let here just -B flag.

        """
        self._exec(server="ripe (no -r)", server_url="whois.ripe.net")  # no -r flag
        return self.get_abusemail()
        # XX
        # if self.abusemail == Config.UNKNOWN_NAME:
        #     self._exec(server="ripe (-B flag)", server_url="whois.ripe.net -B")  # with -B flag
        #     self.get_abusemail(True)

    @staticmethod
    def _str2prefix(s):
        """ Accepts formats:
            88.174.0.0 - 88.187.255.255, 216.245.0.0/18, 2000::/7 ...
        """
        # We have to strip it because of
        #   whois 172.97.38.164
        #   network:netrange:172.97.36.0 -  172.97.39.255
        sp = [a.strip() for a in s.split(" - ")]
        try:
            if len(sp) > 1:
                return IPRange(*sp)
            else:
                return IPNetwork(s)
        except Exception as e:
            logger.warning("Prefix {} cannot be parsed.".format(s))
            Config.error_caught()

    def _match_response(self, patterns, last_word=False):
        """
        :param patterns: pattern string or list of strings
        :param last_word: returns only the last word of whole matched expression else last group (ex: the one in parentheses)

        # , take_nth=None, group=None
        # :param take_nth: if available, return n-th result instead of the first available
        #     I.E. `whois 131.72.138.234 | grep ountr` returns three countries: UY, CL, CL.
        #     ARIN registry informs us that this IP is a LACNIC resource and prints out LACNIC address in UY.
        #     However, CL is the country the IP is hosted in.
        # :param group: returned group - default: last group is returned (ex: the one in parentheses)
        :return:
        """
        if type(patterns) is str:
            patterns = [patterns]

        for chunk in self.whois_response:
            for pattern in patterns:
                # it = re.finditer(pattern, self.whoisResponse) if type(pattern) is str else pattern(self.whoisResponse)
                match = re.search(pattern, chunk)
                # for i, match in enumerate(re.finditer(pattern, chunk)):
                #     if not take_nth or i + 1 == take_nth:
                #         break

                if match:
                    if last_word:  # returns only last word
                        return re.search(r'[^\s]*$', match[0]).group(0)
                    else:
                        return match[len(match.groups())]
        return ""  # no pattern result found

        # for line in self.whoisResponse.split("\n"):
        #     result = re.search(grep, line)
        #     if result:
        #         if lastWord:  # returns only last word
        #             return re.search('[^\s]*$', line).group(0)  # \w*
        #         else:  # returns whole line
        #             return line
        # return ""  # no grep result found

    def analyze(self):
        """
        :return: prefix, "local"|"abroad", incident-contact ( = abuse-mail|country), asn, netname, country, abuse-mail, TTL
        """
        prefix = country = ""

        for server in list(self.servers):
            self._exec(server=server)
            while True:
                # 154.48.234.95
                #   Found a referral to rwhois.cogentco.com:4321.
                #   network:IP-Network:154.48.224.0/19
                #   network:Country:DE
                # 82.175.175.231 'country: NL # BE GB DE LU' -> 'NL'
                country = self._match_response(r'country(-code)?:\s*([a-z]{2})')
                if country == "eu":
                    # "EU # Worldwide" (2a0d:f407:1003::/48)
                    # 'EU # Country is really world wide' (64.9.241.202)
                    # 'EU' (89.41.60.38) (RIPE returned this value)
                    country = ""

                if not country and server == "general":
                    if self._match_response("no match found for n +"):
                        # whois 141.138.197.0/24 ends with this phrase and does not try RIPE which works
                        self._exec(server="ripe", server_url="whois.ripe.net")
                        continue
                    if self._match_response(
                            "the whois is temporary unable to query arin for the requested resource. please try again later"):
                        # whois 154.48.234.95 sometimes ends up like this - when we ask ARIN, we can hang too
                        self._exec(server="arin", server_url="whois.arin.net")
                        continue
                    if self._match_response("block not managed by the ripe ncc"):
                        # whois 109.244.112.0 replies RIPE that they don't manage the block
                        self._exec(server="apnic", server_url="whois.apnic.net")
                        continue
                    if self._match_response("query rate limit exceeded"):  # LACNIC gave me this - seems 300 s needed
                        self.quota.try_start()
                        if Config.get("lacnic_quota_skip_lines", "FIELDS") and not self.slow_mode:
                            if self.see:
                                print("LACNIC quota exceeded.")
                            self.queued_ips.add(self.ip)
                            raise self.quota.QuotaExceeded
                        else:
                            logger.warning(f"Whois server {self.last_server} query rate limit exceeded for: {self.ip}."
                                           f" Sleeping for 300 s till {self.quota.time()}... (you may howevec Ctrl-C to skip)")
                            sleep(300)
                            self._exec(server=server)
                            continue
                    if self.last_server == "rwhois.gin.ntt.net":  # 204.2.250.0
                        self._exec(server="arin", server_url="whois.arin.net")
                        continue
                    if self.last_server in ["whois.twnic.net", "whois.nic.ad.jp", "whois.nic.or.kr"]:
                        # twnic
                        # 210.241.57.0
                        # whois 203.66.23.2 replies whois.twnic.net.tw with "The IP address not belong to TWNIC"
                        # jp: 185.243.43.0
                        # krnic: 125.129.170.2
                        self._exec(server="apnic", server_url="whois.apnic.net")
                        continue

                # loads prefix
                match = self._match_response(["% abuse contact for '([^']*)'",
                                              "% information related to '([^']*)'",
                                              # ip 151.80.121.243 needed this , % information related to
                                              # \'151.80.121.224 - 151.80.121.255\'\n\n% no abuse contact registered
                                              # for 151.80.121.224 - 151.80.121.255
                                              r"inetnum:\s*(.*)",  # inetnum:        151.80.121.224 - 151.80.121.255
                                              r"netrange:\s*(.*)",  # NetRange:       216.245.0.0 - 216.245.63.255
                                              r"cidr:\s*(.*)",  # CIDR:           216.245.0.0/18
                                              r"network:ip-network:\s*(.*)"
                                              # whois 154.48.250.2 "network:IP-Network:154.48.224.0/19"
                                              ])
                if match:
                    prefix = self._str2prefix(match)
                    if prefix and prefix == whole_space:
                        # whois 104.224.36.20 asks -h whois.apnic.net
                        # inetnum:        0.0.0.0 - 255.255.255.255
                        # netname:        IANA-BLOCK
                        # descr:          General placeholder reference for all IPv4 addresses
                        # country:        AU
                        # RIPE if asked does provide at least prefix:
                        # % No abuse contact registered for 104.166.192.0 - 104.232.35.255
                        #
                        # inetnum:        104.166.192.0 - 104.232.35.255
                        # netname:        NON-RIPE-NCC-MANAGED-ADDRESS-BLOCK
                        prefix = None
                        if server == "general":
                            self._exec(server="ripe", server_url="whois.ripe.net")
                            server = "disabled apnic"
                            continue

                if not country:
                    country = self._load_country_from_addresses()
                break
            if not country:
                fail = None
                if self._match_response("network is unreachable") or self._match_response("name or service not known"):
                    fail = f"Whois server {self.servers[server]} is unreachable. Disabling for this session."
                if self._match_response("access denied"):  # RIPE gave me this
                    fail = f"Whois server {self.servers[server]} access denied. Disabling for this session."
                if self._match_response("invalid search key"):
                    logger.warning(f"Invalid search key for: {self.ip}")

                if fail:
                    logger.warning(fail)
                    Whois.servers.pop(server)
                    continue
            else:
                break

        asn = self._match_response(r'\norigin(.*)\d+', last_word=True)
        netname = self._match_response([r'netname:\s*([^\s]*)', r'network:network-name:\s*([^\s]*)'])

        ab = self.get_abusemail()
        if Whois.unknown_mode and not ab:
            ab = self.resolve_unknown_mail()

        local = Config.get("local_country", "FIELDS")
        if local and country not in local:
            mail = Contacts.country2mail[country] if country in Contacts.country2mail else ab
            get1 = "abroad"
            get2 = f"{country}{Config.ABROAD_MARK}{mail}" if mail else ""
        else:
            get1 = "local"
            get2 = ab
        return prefix, get1, get2, asn, netname, country, ab, int(time())

    def _load_country_from_addresses(self):
        # let's try to find country in the non-standardised address field
        for address in re.findall(r"address:\s+(.*)", "\n".join(self.whois_response)):
            c = address_country_lowered(address)
            if c:
                logger.info(f"Found country in {address}")
                return c
        return ""

    reAbuse = re.compile('[a-z0-9._%+-]{1,64}@(?:[a-z0-9-]{1,63}\.){1,125}[a-z]{2,63}')

    def get_abusemail(self):
        """ Loads abusemail from last whois response OR from whois json api. """
        match = self.reAbuse.search(self._match_response(['% abuse contact for.*',
                                                          'orgabuseemail.*',
                                                          'abuse-mailbox.*',
                                                          "e-mail:.*"  # whois 179.50.80.0/21
                                                          ]))
        return match.group(0) if match else ""

    regRe = re.compile(r"using server (.*)\.")

    def _exec(self, server, server_url=None):
        """ Query whois server """
        if server == "general":
            cmd = ["whois", "--verbose", self.ip]
        else:
            if not server_url:
                server_url = Whois.servers[server]
            cmd = ["whois", "--verbose", "-h", server_url, "--", self.ip]
        self.last_server = None  # check what registry whois asks - may use a strange LIR that returns non-senses
        try:
            # in case wrong env is set to whois, we get `147.32.106.205` country NL and not CZ
            # because we will not find string "found a referral to " in the WHOIS response
            p = Popen(cmd, shell=False, stdin=PIPE, stdout=PIPE, stderr=PIPE, env=subprocess_env)
            response = p.stdout.read().decode("unicode_escape").strip().lower()  # .replace("\n", " ")
            response += p.stderr.read().decode("unicode_escape").strip().lower()
        except UnicodeDecodeError:
            # ip address 94.230.155.109 had this string 'Jan Krivsky Hl\xc3\x83\x83\xc3\x82\xc2\xa1dkov' and everything failed
            self.whois_response = []
            logger.warning("Whois response for IP {} on server {} cannot be parsed.".format(self.ip, server))
        except TypeError:  # could not resolve host
            self.whois_response = []
        except FileNotFoundError:
            Config.missing_dependency("whois")
        else:
            try:
                self.last_server = Whois.regRe.search(response).groups()[0]
            except (IndexError, AttributeError):
                pass

            # Sometimes, a registry calls another registry for you. This may chain.
            # We prioritize by the most recent to the first.
            # So when whois 154.48.234.95 goes to AfriNIC that goes to ARIN that goes to rwhois.cogentco.com:4321,
            # we find country in Cogento, then in ARIN, then in AfriNIC.
            # This may lead to the behaviour when Country is from Cogento and Netname is from ARIN.
            # I don't know how to handle this better.
            # Another example is
            # whois 198.55.103.47 leads to "Found a referral to rwhois.quadranet.com:4321."
            # 154.48.234.95 goes to AfriNIC that goes to ARIN that says:
            #   CIDR:           154.48.0.0/16
            #   Country:        US
            #   Found a referral to rwhois.cogentco.com:4321.
            #   network:IP-Network:154.48.224.0/19
            #   network:Country:DE
            ref_s = "found a referral to "
            self.whois_response = response.split(ref_s)[::-1]

            # i = self.whoisResponse.find(ref_s)
            # if i > -1:
            #     self.whoisResponse = self.whoisResponse[i + len(ref_s):]
        finally:
            Whois.stats[self.last_server or server] += 1



import csv
import re
from difflib import SequenceMatcher
import logging
from typing import Any, Callable, List, Optional
from pathlib import Path
from sys import exit

from dialog import Dialog
from prompt_toolkit.shortcuts import clear

from .action import AggregateAction, MergeAction
from .config import Config
from .decorators import PickBase, PickMethod, PickInput
from .dialogue import Cancelled, Menu, choose_file, csv_split, hit_any_key, pick_option, ask, is_yes
from .field import Field
from .parser import Parser
from .types import Types, TypeGroup, types, Type, graph, methods, Aggregate, get_module_from_path
from .wizzard import Preview
from .wrapper import Wrapper

logger = logging.getLogger(__name__)
aggregate_functions = [f for f in Aggregate.__dict__ if not f.startswith("_")]
aggregate_functions_str = "".join("\n* " + f for f in aggregate_functions)

class ActionController:
    def __init__(self, parser: Parser, reprocess = False):
        self.parser = parser
        self.reprocess = reprocess

    def add_column(self):
        self.select_col("New column", only_computables=True, add=True)
        self.parser.is_processable = True

    def add_aggregation(self, fn_name:str = None, column_task:Optional[str] = None, group:Optional[Field]=None, grouping_probably_wanted=True, exit_on_fail=False):
        # choose what column we want

        fn = self.get_aggregation_fn(fn_name, exit_on_fail)
        if column_task:
            field: Field = self.parser.fields[self.parser.identifier.get_column_i(column_task, "to be aggregated with")]
        else:
            field = self.select_col("aggregation")

        sett = self.parser.settings["aggregate"]
        if sett:
            group_old, fns = sett.group_by, sett.actions
        else:
            group_old, fns = None, []

        group = self.assure_aggregation_group_by(fn, field, group or group_old, grouping_probably_wanted, exit_on_fail)
        fns.append([fn, field])
        self.parser.settings["aggregate"] = AggregateAction(group, fns)
        self.parser.is_processable = True

    def add_dialect(self):
        # XX not ideal and mostly copies Parser.__init__
        # XX There might be a table with all the csv.dialect properties or so.
        dialect = self.parser.settings["dialect"]
        while True:
            s = "What is delimiter " + (f"(default '{dialect.delimiter}')" if dialect.delimiter else "") + ": "
            dialect.delimiter = input(s) or dialect.delimiter
            if len(dialect.delimiter) != 1:
                print("Delimiter must be a 1-character string. Invent one (like ',').")
                continue
            s = "What is quoting char " + (f"(default '{dialect.quotechar}')" if dialect.quotechar else "") + ": "
            dialect.quotechar = input(s) or dialect.quotechar
            break
        dialect.quoting = csv.QUOTE_NONE if not dialect.quotechar else csv.QUOTE_MINIMAL

        if self.parser.has_header:
            if self.parser.settings['header'] is False:
                s = f"Should we remove header? "
            else:
                s = f"Should we include header? "
            if not is_yes(s):
                self.parser.settings["header"] = not self.parser.settings["header"]

        self.parser.is_processable = True



    def add_filter(self):
        menu = Menu(title="Choose a filter")
        menu.add("Unique filter", self.add_uniquing)
        menu.add("Include filter", self.add_filtering)
        menu.add("Exclude filter", lambda: self.add_filtering(False))
        menu.sout()

    def add_merge(self, remote_path=None, remote_col_i:Optional[int]=None, local_col_i:Optional[int]=None):
        if not remote_path:
            remote_path = choose_file("What file we should merge the columns from?")
        wrapper2 = Wrapper(Path(remote_path), reprocess=self.reprocess)
        parser2 = wrapper2.parser

        # dialog user and build the link between files
        controller2 = ActionController(parser2)
        column1 = self.parser.fields[local_col_i] if local_col_i is not None else None

        if remote_col_i is not None:
            column2 = parser2.fields[remote_col_i]
        else:
            high = parser2.get_similar(column1 or self.parser.fields)
            column2 = controller2.select_col("Select remote column to be merged", include_computables=False, highlighted=high)
        if not column1:
            high = self.parser.get_similar(column2)
            column1 = self.select_col(f"Select local column to merge '{column2}' to", include_computables=False, highlighted=high)

        # cache remote values
        operation = MergeAction.build(wrapper2.file, parser2, column2, column1)

        # build local fields based on the remotes
        for rf in parser2.fields:
            f = Field(rf.name,
                      is_chosen=False if rf is column2 else True,
                      merged_from=rf,
                      merge_operation=operation)
            self.parser.add_field(append=f)

        # prepare the operation
        self.parser.is_processable = True
        self.parser.settings["merge"].append(operation)

    def add_new_column(self, task, add=True):
        """
        :type task: str FIELD,[COLUMN],[SOURCE_TYPE], ex: `netname 3|"IP address" ip|sourceip`
        :type add: bool Add to the result.
        """
        target_type, source_field, source_type, custom = self._add_new_column(task)
        if not self.source_new_column(target_type, add, source_field, source_type, custom):
            print("Cancelled")
            exit()
        self.parser.is_processable = True
        return target_type

    def _add_new_column(self, task):
        """ Internal analysis for `add_new_column` """
        task = csv_split(task)
        custom = []
        target_type = task[0]
        m = re.search(r"(\w*)\[([^]]*)\]", target_type)
        if m:
            target_type = m.group(1)
            custom = [m.group(2)]
        try:
            target_type = types[types.index(target_type)]  # determine FIELD by exact name
        except ValueError:
            d = {t.name: SequenceMatcher(None, task[0], t.name).ratio() for t in Types.get_computable_types()}
            rather = max(d, key=d.get)
            logger.error(f"Unknown field '{task[0]}', did not you mean '{rather}'?")
            exit()
        source_field, source_type, c = self.parser.identifier.get_fitting_source(target_type, *task[1:])
        custom = c + custom
        return target_type, source_field, source_type, custom


    def add_uniquing(self, col_i=None):
        if col_i is None:
            col_i = self.select_col("unique").col_i
        self.parser.settings["unique"].append(col_i)
        self.parser.is_processable = True

    def assure_aggregation_group_by(self, fn, field, group, grouping_probably_wanted=True, exit_on_fail=False) -> Optional[Field]:
        a,b,c = (fn == Aggregate.count, group is None, grouping_probably_wanted)
        if (a,b) == (True, True):
            group = field
        elif (a,b,c) == (False, True, True):
            # here, self.select col might return None
            group = self.select_col("group by", prepended_field=("no grouping", "aggregate whole column"))
        elif (a,b) == (True, False):
            if field != group:
                logger.error(f"Count column '{field.name}' must be the same"
                                f" as the grouping column '{group.name}'.")
                if exit_on_fail:
                    exit()
                else:
                    raise Cancelled
        return group # XX as of Python 3.10 replace with the following
        # match (fn == Aggregate.count, group is None, grouping_probably_wanted):
        #     case True, True, _:
        #         group = field
        #     case False, True, True:
        #         # here, self.select col might return None
        #         group = self.select_col("group by", prepended_field=("no grouping", "aggregate whole column"))
        #     case True, False, _:
        #         if field != group:
        #             logger.error(f"Count column '{field.name}' must be the same"
        #                             f" as the grouping column '{group.name}'.")
        #             if exit_on_fail:
        #                 exit()
        #             else:
        #                 raise Cancelled
        # return group





    def add_filtering(self, include=True, col_i=None, val=None):
        if col_i is None:
            col_i = self.select_col("filter").col_i
        if val is None:
            s = "" if include else "not "
            val = ask(f"What value must {s}the field have to keep the line?")
        self.parser.settings["filter"].append((include, col_i, val))
        self.parser.is_processable = True

    def add_splitting(self):
        self.parser.settings["split"] = self.select_col("splitting").col_i
        self.parser.is_processable = True


    def choose_cols(self):
        # XX possibility un/check all
        chosens = [(str(i + 1), str(f), f.is_chosen) for i, f in enumerate(self.parser.fields)]
        d = Dialog(autowidgetsize=True)
        ret, values = d.checklist("What fields should be included in the output file?", choices=chosens)
        if ret == "ok":
            for f in self.parser.fields:
                f.is_chosen = False
            for v in values:
                self.parser.fields[int(v) - 1].is_chosen = True
            self.parser.is_processable = True


    def get_aggregation_fn(self, fn_name:str=None, exit_on_fail=False) -> Optional[Callable]:
        if not fn_name:
            menu = Menu("Choose aggregate function", callbacks=False, fullscreen=True)
            [menu.add(f) for f in aggregate_functions]
            option = menu.sout()
            if not option:
                raise Cancelled # XXX tohle bude fungovat?
            fn_name = aggregate_functions[int(option) - 1]
        fn = getattr(Aggregate, fn_name, None)
        if not fn:
            if exit_on_fail:
                logger.error(f"Unknown aggregate function '{fn_name}'. Possible functions are: {aggregate_functions_str}")
                exit()
            else:
                raise Cancelled
        return fn


    def select_col(self, dialog_title="", only_computables=False, include_computables=True, add=None, prepended_field=None, highlighted: Optional[List[Field]] = None) -> Optional[Field]:
        """ Starts dialog where user has to choose a column.
            If cancelled, we return to main menu automatically.
            :type prepended_field: tuple (field_name, description) If present, this field is prepended. If chosen, you receive None.
        """
        # add existing fields
        fields = [] if only_computables else self.parser.get_fields_autodetection()

        # add computable field types
        if include_computables:
            for field in Types.get_computable_types():
                if field.from_message:
                    s = field.from_message
                else:
                    node_distance = graph.dijkstra(field, ignore_private=True)
                    s = field.group.name + " " if field.group != TypeGroup.general else ""
                    s += "from " + ", ".join([str(k) for k in node_distance][:3])
                    if len(node_distance) > 3:
                        s += "..."
                fields.append((f"new {field}...", s))

        # add special prepended field
        if prepended_field:
            fields.insert(0, prepended_field)

        # launch dialog
        col_i = pick_option(fields, dialog_title, guesses=[f.col_i for f in highlighted or ()])

        # convert returned int col_i to match an existing or new column
        if prepended_field:
            col_i -= 1
            if col_i == -1: # we hit a prepended_field, a mere description, not a real field
                return None
        if only_computables or col_i >= len(self.parser.fields):
            target_type_i = col_i if only_computables else col_i - len(self.parser.fields)
            col_i = len(self.parser.fields)
            self.source_new_column(Types.get_computable_types()[target_type_i], add=add)

        return self.parser.fields[col_i]



    def source_new_column(self, target_type, add=None, source_field: Field = None, source_type: Type = None,
                          custom: list = None):
        """ We know what Field the new column should be of, now determine how we should extend it:
            Summarize what order has the source field and what type the source field should be considered alike.
                :type source_field: Field
                :type source_type: Type
                :type target_type: Type
                :type add: bool if the column should be added to the table; None ask
                :type custom: List
                :raise Cancelled
                :return Field
        """
        if custom is None:
            # default [] would be evaluated at the time the function is defined, multiple columns may share the same function
            custom = []
        dialog = Dialog(autowidgetsize=True)
        if not source_field or not source_type:
            print("\nWhat column we base {} on?".format(target_type))
            guesses = self.parser.identifier.get_fitting_source_i(target_type)
            source_col_i = pick_option(self.parser.get_fields_autodetection(),
                                       title="Searching source for " + str(target_type),
                                       guesses=guesses)
            source_field = self.parser.fields[source_col_i]
            source_type = self.parser.identifier.get_fitting_type(source_col_i, target_type, try_plaintext=True)
            if source_type is None:
                # ask how should be treated the column as, even it seems not valid
                # list all known methods to compute the desired new_field (e.g. for incident-contact it is: ip, hostname, ...)
                choices = [(k.name, k.description)
                           for k, _ in graph.dijkstra(target_type, ignore_private=True).items()]

                # if len(choices) == 1 and choices[0][0] == Types.plaintext:
                #     # if the only method is derivable from a plaintext, no worries that a method
                #     # converting the column type to "plaintext" is not defined; everything's plaintext
                #     source_type = Types.plaintext
                # el
                if choices:
                    s = ""
                    if self.parser.sample_parsed:
                        s = f"\n\nWhat type of value '{self.parser.sample_parsed[0][source_col_i]}' is?"
                    title = f"Choose the right method\n\nNo known method for making {target_type} from column {source_field} because the column type wasn't identified. How should I treat the column?{s}"
                    code, source_type = dialog.menu(title, choices=choices)
                    if code == "cancel":
                        raise Cancelled("... cancelled")
                    source_type = getattr(Types, source_type)
                else:
                    dialog.msgbox(
                        "No known method for making {}. Raise your usecase as an issue at {}.".format(target_type,
                                                                                                      Config.PROJECT_SITE))
                    raise Cancelled("... cancelled")
            clear()

        if not custom:
            try:
                if target_type.group == TypeGroup.custom:
                    if target_type == Types.code:
                        print("What code should be executed? Change 'x'. Ex: x += \"append\";")
                        custom = Preview(source_field, source_type).code()
                    elif target_type in [Types.reg, Types.reg_m, Types.reg_s]:
                        *custom, target_type = Preview(source_field, source_type, target_type).reg()
                    elif target_type == Types.external:  # choose a file with a needed method
                        while True:
                            path = choose_file("What .py file should be used as custom source?", dialog)
                            module = get_module_from_path(path)
                            if module:
                                # inspect the .py file, extract methods and let the user choose one
                                code, method_name = dialog.menu(f"What method should be used in the file {path}?",
                                                                choices=[(x, "") for x in dir(module) if
                                                                         not x.startswith("_")])
                                if code == "cancel":
                                    raise Cancelled("... cancelled")

                                custom = path, method_name
                                break
                            else:
                                dialog.msgbox("The file {} does not exist or is not a valid .py file.".format(path))
                    if not custom:
                        raise Cancelled("... cancelled")
            except Cancelled:
                raise Cancelled("... cancelled")
            path = graph.dijkstra(target_type, start=source_type, ignore_private=True)
            for i in range(len(path) - 1):
                m = methods[path[i], path[i + 1]]
                if isinstance(m, PickBase):
                    c = None
                    if Config.get("yes"):
                        pass
                    elif type(m) is PickMethod:
                        m: PickMethod
                        code, c = dialog.menu(f"Choose subtype", choices=m.get_options())
                        if code == "cancel":
                            raise Cancelled("... cancelled")
                    elif type(m) is PickInput:
                        m: PickInput
                        c = Preview(source_field, source_type, target_type).pick_input(m)
                    custom.insert(0, c)
        if add is None:
            if dialog.yesno("New field added: {}\n\nDo you want to include this field as a new column?".format(
                    target_type)) == "ok":
                add = True

        f = Field(target_type, is_chosen=add,
                  source_field=source_field,
                  source_type=source_type,
                  new_custom=custom)
        if f.source_field.merged_from:
            hit_any_key("Unfortunately, sourcing from columns being merged was not implemented."
                        " Ask for it on Github. And for now, merge first, then add a column.")
            raise Cancelled("Sourcing from columns being merged was not implemented")
        self.parser.settings["add"].append(f)
        self.parser.add_field(append=f)
        return f


import logging
import os
import socket
import subprocess
import sys
from ast import literal_eval
from io import StringIO
from pathlib import Path
from sys import exit
from tempfile import NamedTemporaryFile
from typing import Optional, TYPE_CHECKING

from colorama import init as colorama_init, Fore
from dialog import Dialog
from prompt_toolkit import PromptSession, HTML
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.shortcuts import clear
from validate_email import validate_email

from .action import AggregateAction
from .action_controller import ActionController, aggregate_functions
from .args_controller import ArgsController, otrs_flags, new_fields
from .attachment import Attachment
from .config import Config, console_handler, edit, get_path
from .contacts import Contacts
from .flag import MergeFlag, FlagController
from .decorators import PickBase, PickMethod, PickInput
from .dialogue import Cancelled, Debugged, Menu, csv_split, pick_option, ask, ask_number
from .field import Field
from .ipc import socket_file, recv, send, daemon_pid
from .mail_sender import MailSenderOtrs, MailSenderSmtp
from .parser import Parser
from .types import Types, TypeGroup, Aggregate
from .wizzard import bottom_plain_style
from .wrapper import Wrapper
from . import __version__

if TYPE_CHECKING:
    from socket import socket

logger = logging.getLogger(__name__)

def send_ipc(pipe, msg, refresh_stdout):
    sys.stdout = sys.stderr = StringIO()  # creating a new one is faster than truncating the old
    console_handler.setStream(sys.stdout)
    sys.stdout_real.write(refresh_stdout)
    return send(pipe, msg)


def control_daemon(cmd, in_daemon=False):
    if cmd == "stop":
        Config.set("daemonize", False)  # daemon should not be started at the end
        print("Convey daemon stopping.")
    if cmd in ["restart", "start"]:
        if in_daemon:
            raise ConnectionResetError("restart.")
        else:
            Config.set("daemonize", True)
            control_daemon("kill")
        if cmd == "start":
            print("Convey daemon starting.")
            exit()
    if cmd in ["kill", "stop"]:
        if in_daemon:
            raise ConnectionResetError("stop.")
        else:
            pid = daemon_pid()
            if pid:
                subprocess.run(["kill", pid])
            if Path(socket_file).exists():
                Path(socket_file).unlink()
    if cmd == "stop":
        exit()
    if cmd == "status":
        if daemon_pid():
            print(f"Convey daemon is listening at socket {socket_file}")
        else:
            print("Convey daemon seems not to be running.")
        exit()
    if cmd is False:
        if in_daemon:
            raise ConnectionAbortedError("Daemon should not be used")
        else:
            Config.set("daemonize", False)
    if cmd == "server":
        if in_daemon:
            raise ConnectionAbortedError("The new process will become the new server.")
        else:
            control_daemon("kill")
    return cmd


class Controller:
    def __init__(self, parser=None):
        self.parser = parser

    def run(self, given_args: Optional[str] = None):
        """ given_args - Input for ArgumentParser. Else sys.argv used. """
        if "--disable-external" in sys.argv:
            Config.set("disable_external", True)
        Types.refresh()  # load types so that we can print out computable types in the help text
        argparser = ArgsController().parse_args()
        self.args = args = argparser.parse_args(args=given_args)
        self.see_menu = True
        self.check_server(args)
        is_daemon, stdout, server = self.check_daemon(args)
        self.assure_terminal(is_daemon, args)
        ac = self.process_args_from_daemon_or_locally(is_daemon, stdout, server, argparser, args)
        self.run_menu(ac)

    def check_server(self, args):
        if args.server:
            # XX not implemeneted: allow or disable fields via CLI by ex: `--web`
            print(f"Webserver configuration can be changed by `convey --config uwsgi`")
            print(get_path("uwsgi.ini").read_text())
            cmd = ["uwsgi", "--ini", get_path("uwsgi.ini"), "--wsgi-file", Path(Path(__file__).parent, "wsgi.py")]
            subprocess.run(cmd)
            exit()

    def check_daemon(self, args):
        if args.daemon is not None and control_daemon(args.daemon) == "server":
            # XX :( after a thousand requests, we start to slow down. Memory leak must be somewhere
            Config.set("daemonize",
                       False)  # do not restart daemon when killed, there must be a reason this daemon was killed
            if Path(socket_file).exists():
                Path(socket_file).unlink()

            try:
                Config.init_verbosity(args.yes, 30 if args.quiet else (10 if args.verbose else None), True)
                Config.integrity_check()
            except ConnectionAbortedError:
                print("Config file integrity check failed. Launch convey normally to upgrade config parameters.")
                exit()

            print("Opening socket...")
            server = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            server.bind(socket_file)
            server.listen()
            sys.stdout_real = stdout = sys.stdout
            sys.stdout = sys.stderr = StringIO()
            if sys.version_info >= (3, 7):  # XX remove when dropped Python 3.6 support.
                # In 3.6, logging message from the daemon will not work.
                console_handler.setStream(sys.stdout)
            PromptSession.__init__ = lambda _, *ar, **kw: (_ for _ in ()).throw(
                ConnectionAbortedError('Prompt raised.'))
            return True, stdout, server
        return None, None, None

    def assure_terminal(self, is_daemon, args):
        if not is_daemon and not sys.stdin.isatty():  # piping to the process, no terminal
            try:
                sys.stdin = open("/dev/tty")
            except OSError:
                # this might work on Windows platform
                # Return "" for every input we got.
                PromptSession.prompt = lambda *ar, **kw: ""
            else:
                # Let's make a safe_prompt that is able to read from /dev/tty.
                # prompt_toolkit does not work with /dev/tty stdin
                # because it is a POSIX pipe only and it needs a "pseudo terminal pipe"
                # see https://github.com/prompt-toolkit/python-prompt-toolkit/issues/502
                # and menu does not work good
                def safe_prompt_2(*ar, **kw):
                    return input(ar[1] if len(ar) > 1 else kw["message"] if "message" in kw else "?")

                def safe_prompt(*ar, **kw):
                    print(
                        "This is just an emergency input mode because convey interactivity works bad when piping into.")
                    PromptSession.prompt = safe_prompt_2
                    return safe_prompt_2(*ar, **kw)

                PromptSession.prompt = safe_prompt

            self.see_menu = False
            args.yes = True

    def process_args_from_daemon_or_locally(self, is_daemon, stdout, server, argparser, args):
        colorama_init()
        while True:
            try:
                if is_daemon:
                    stdout.write("Listening...\n")
                    stdout.flush()
                    pipe, _addr = server.accept()
                    msg = recv(pipe)
                    if not msg:
                        continue
                    stdout.write("Accepted: " + msg + "\n")
                    stdout.flush()
                    argv = literal_eval(msg)
                    if not argv:
                        raise ConnectionAbortedError("No arguments passed")
                    elif "--disable-external" in argv:
                        # Param "disable_external" is parsed before the help text normally
                        # but we are in daemon and the help text has already been prepared with externals.
                        raise ConnectionAbortedError("Disable externals might change the behaviour (help text, ...)")
                    try:
                        os.chdir(Path(argv[0]))
                    except OSError:
                        stdout.write("Invalid cwd\n")
                        continue
                    self.cleanup()  # reset new fields so that they will not be remembered in another query
                    try:
                        self.args = args = argparser.parse_args(argv[2:])  # the daemon has receives a new command
                    except SystemExit as e:
                        if not sys.stdout.getvalue():
                            # argparse sent usage to stderr, we do not have in in stdout instance will rerun the command
                            raise ConnectionAbortedError("Error in help text")
                        else:
                            # argparse put everything in stdout
                            exit()
                    self.see_menu = True
                    control_daemon(args.daemon, True)
                ac = self.process_args(is_daemon, args, argparser)
                if not ac:
                    continue
            except ConnectionRefusedError as e:
                send_ipc(pipe, chr(3), f"Daemon has insufficient input: {e}\n")
                continue
            except ConnectionAbortedError as e:
                send_ipc(pipe, chr(4), "Daemon cannot help: " + (str(e) or "Probably a user dialog is needed.") + "\n")
                continue
            except ConnectionResetError as e:
                send_ipc(pipe, chr(17), f"Daemon killed: {e}\n")
                exit()
            except SystemExit:
                if is_daemon:
                    send_ipc(pipe, sys.stdout.getvalue(), "Result sent.\n")
                    continue  # wait for next IPC connection
                else:
                    raise
            break
        return ac

    def process_args(self, is_daemon, args, argparser) -> Optional[ActionController]:
        # this try-block may send the results to the client convey processes when a daemon is used
        if args.server:
            raise ConnectionAbortedError("web server request")
        if args.config is not None:
            edit(*args.config, restart_when_done=True)
            exit()
        Config.set("stdout", args.output is True or None)
        if args.output is True:
            # --output=True → no output file in favour of stdout (Config.get("stdout") -> parser.stdout set)
            # --output=FILE → an output file generated (Config.get("output") -> parser.target_file set)
            args.output = None
        for flag in ["output", "web", "whois", "nmap", "dig", "delimiter", "quote_char", "compute_preview",
                        "user_agent",
                        "multiple_hostname_ip", "multiple_cidr_ip", "web_timeout", "whois_ttl", "disable_external",
                        "debug", "crash_post_mortem",
                        "testing", "attach_files", "attach_paths_from_path_column", "jinja", "subject", "body", "references",
                        "whois_delete_unknown", "whois_reprocessable_unknown", "whois_cache"]:
            if getattr(args, flag) is not None:
                Config.set(flag, getattr(args, flag))
        if args.headless or args.send_test:
            args.yes = True
            args.quiet = True
            self.see_menu = False
        Config.init_verbosity(args.yes, 30 if args.quiet else (10 if args.verbose else None), is_daemon)
        if is_daemon:
            logger.debug("This result comes from the daemon.")
        Types.refresh()  # reload Types for the second time so that the methods reflect CLI flags
        TypeGroup.init()
        if args.show_uml is not None:
            print(Types.get_uml(args.show_uml))
            exit()
        if args.get_autocompletion:
            print(self.get_autocompletion(argparser))
            exit()
        if args.version:
            print(__version__)
            exit()
        if not is_daemon:  # in daemon, we checked it earlier
            Config.integrity_check()
        if args.header:
            Config.set("header", True)
        if args.no_header:
            Config.set("header", False)
        if args.csv_processing:
            Config.set("single_query", False)
        if args.single_query or args.single_detect:
            Config.set("single_query", True)
            if args.single_detect:
                Config.set("single_detect", True)
        Config.set("adding-new-fields", bool(new_fields))
        self.wrapper = Wrapper(args.file_or_input, args.file, args.input,
                                args.type, args.fresh, args.reprocess,
                                args.whois_delete)
        self.parser: Parser = self.wrapper.parser
        ac = ActionController(self.parser, args.reprocess)

        if args.threads is not None:
            Config.set("threads", args.threads)

        def get_column_i(col, check):
            self.parser.is_processable = True
            return self.parser.identifier.get_column_i(col, check=check)

        # load flags
        for flag in otrs_flags:
            if args.__dict__[flag[0]]:
                self.parser.__dict__[flag[0]] = args.__dict__[flag[0]]
                logger.debug("{}: {}".format(flag[1], flag[0]))

        # prepare some columns to be removed
        if args.delete:
            for c in args.delete.split(","):
                self.parser.fields[get_column_i(c, "to be deleted")].is_chosen = False

        # append new fields from CLI
        for add, task in new_fields:
            ac.add_new_column(task, add)

        # merge
        fc = FlagController(self.parser)
        if args.merge:
            ac.add_merge(**fc.read(MergeFlag, args.merge))

        # run single value check if the input is not a CSV file
        if args.single_detect:
            exit()
        if self.parser.is_single_query:
            res = self.parser.run_single_query(json=args.json)
            if res:
                print(res)
            exit()
        if is_daemon and self.see_menu: # if we will need menu, daemon must stop here
            raise ConnectionAbortedError("displaying a menu is too complex")

        if args.aggregate:
            params = csv_split(args.aggregate)
            group: Optional[Field] = self.parser.fields[get_column_i(params.pop(), "to be grouped by")] \
                if len(params) % 2 else None
            if not params:
                ac.add_aggregation(Aggregate.count.__name__, group, exit_on_fail=True)
            else:
                for i in range(0, len(params), 2):
                    column_task, fn_name = params[i:i + 2]
                    ac.add_aggregation(fn_name, column_task, group, grouping_probably_wanted=False, exit_on_fail=True)

        if args.sort:
            self.parser.resort(csv_split(args.sort))
            self.parser.is_processable = True

        if args.split:
            self.parser.settings["split"] = get_column_i(args.split, "to be split with")

        if args.include_filter:
            col, val = csv_split(args.include_filter)
            ac.add_filtering(True, get_column_i(col, "to be filtered with"), val)

        if args.exclude_filter:
            col, val = csv_split(args.exclude_filter)
            ac.add_filtering(False, get_column_i(col, "to be filtered with"), val)

        if args.unique:
            ac.add_uniquing(get_column_i(args.unique, "to be put a single time"))

        # set output dialect
        # Taken from config.ini or flags. If differs from the current one, parser marked as processable.
        self.parser.settings["dialect"] = dialect = type('', (), {})()
        for k, v in self.parser.dialect.__dict__.copy().items():
            setattr(dialect, k, v)

        def change_dialect(s, s2):
            # delimiter_output, quote_char_output
            v = getattr(args, s + "_output") or Config.get(s + "_output", "CSV")
            if v and v != getattr(dialect, s2):
                v = v.replace(r"\t", "\t").replace("TAB", "\t").replace("tab", "\t")
                if len(v) != 1:
                    print(f"Output {s2} has to be 1 character long: {v}")
                    exit()
                setattr(dialect, s2, v)
                self.parser.is_processable = True

        change_dialect("delimiter", "delimiter")
        change_dialect("quote_char", "quotechar")
        self.parser.settings["header"] = self.parser.has_header
        if self.parser.has_header and args.header_output is not None:
            # If current parser has header, we may cut it off
            # However, this is such a small change, we will not turning parser.is_processable on.
            self.parser.settings["header"] = args.header_output

        if self.parser.is_processable and Config.get("yes"):
            self.process()

        if args.send and self.parser.is_analyzed and self.parser.is_split and not self.parser.is_processable:
            # Config.set("yes", True)
            self.see_menu = False
            if args.send is not True:
                self.send_menu(args.send, send_now=True)
            else:
                self.send_menu(send_now=True)
        if args.send_test:
            c = Path(args.send_test[1]).read_text()
            Path(Config.get_cache_dir(), Config.get("mail_template")).write_text(c)
            Path(Config.get_cache_dir(), Config.get("mail_template_abroad")).write_text(c)
            self.send_menu(test_attachment=args.send_test[0])

        if not self.see_menu:
            self.close()
        if is_daemon:  # if in daemon, everything important has been already sent to STDOUT
            exit()
        return ac

    def run_menu(self, ac):
        # main menu
        start_debugger = False
        session = None
        while True:
            if session and hasattr(session, "process"):
                # session.prompt keybinding asked processing
                # (we cannot reprocess from keybinding due to the deadlock if an input had been encountered,
                # no clear way to call a prompt within another prompt)
                self.process()
            self.parser = self.wrapper.parser  # may be changed by reprocessing
            self.parser.informer.sout_info()

            if start_debugger:
                print("\nDebugging mode, you may want to see `self.parser` variable:")
                start_debugger = False
                Config.get_debugger().set_trace()

            menu = Menu(title="Main menu - how the file should be processed?")
            menu.add("Pick or delete columns", ac.choose_cols)
            menu.add("Add a column", ac.add_column)
            menu.add("Filter", ac.add_filter)
            menu.add("Split by a column", ac.add_splitting)
            menu.add("Change CSV dialect", ac.add_dialect)
            menu.add("Aggregate", ac.add_aggregation)
            menu.add("Merge", ac.add_merge)
            if self.parser.is_processable:
                menu.add("process", self.process, key="p", default=True)
            else:
                menu.add("process  (choose some actions)")
            if self.parser.is_analyzed and self.parser.is_split:
                if self.parser.is_processed:
                    menu.add("send...", self.send_menu, key="s", default=not self.parser.is_processable)
                else:
                    menu.add("send (process first)")
            else:
                menu.add("send (split first)")
            if self.parser.is_analyzed:
                menu.add("show all details", lambda: (self.parser.informer.sout_info(full=True), input()), key="d")
            else:
                menu.add("show all details (process first)")
            menu.add("redo...", self.redo_menu, key="r")
            menu.add("config...", self.config_menu, key="c")
            menu.add("exit", self.close, key="x")

            try:
                bindings = KeyBindings()
                session = PromptSession()

                def refresh():
                    # exiting the app makes the main menu redraw again
                    session.app.exit(session.layout.current_buffer.text or "refresh")

                @bindings.add('right')  # select column
                def _(_):
                    self.parser.move_selection(1)
                    refresh()

                @bindings.add('left')  # select column
                def _(_):
                    self.parser.move_selection(-1)
                    refresh()

                @bindings.add('c-right')  # control-right to move the column
                def _(_):
                    self.parser.move_selection(1, True)
                    refresh()

                @bindings.add('c-left')  # control-left to move the column
                def _(_):
                    self.parser.move_selection(-1, True)
                    refresh()

                @bindings.add('delete')  # enter to toggle selected field
                def _(_):
                    [f.toggle_chosen() for f in self.parser.fields if f.is_selected]
                    refresh()

                @bindings.add('escape', 'a')  # alt-a to aggregate
                def _(_):
                    for f in self.parser.fields:
                        if f.is_selected:
                            self.parser.settings["aggregate"] = AggregateAction(f, [(Aggregate.count, f)])
                            self.parser.is_processable = True
                            session.process = True
                            break
                    else:
                        # I cannot use a mere `input()` here, it would interfere with promtpt_toolkit and freeze
                        Dialog(autowidgetsize=True).msgbox("No column selected to aggregate with.\nUse arrows to select a column first.")
                    refresh()

                # @bindings.add('escape', 'n')  # alt-n to rename header
                # def _(_):
                #     for f in self.parser.fields:
                #         if f.is_selected:
                #             break
                #     refresh()

                options = {'key_bindings': bindings,
                           "bottom_toolbar": HTML("Ctrl+<b>←/→</b> arrows for column manipulation,"
                                                  " <b>Delete</b> for exclusion,"
                                                  " <b>Alt+a</b> to aggregate count"),
                           "style": bottom_plain_style
                           }
                menu.sout(session, options)
            except Cancelled as e:
                print(e)
                pass
            except Debugged as e:
                start_debugger = True

    def send_menu(self, method="smtp", test_attachment=None, send_now=False):
        # choose method SMTP/OTRS
        # We prefer OTRS sending over SMTP because of the signing keys that an OTRS operator does not possess.
        if Config.get("otrs_enabled", "OTRS") and self.args.otrs_id:
            method = "otrs"
        elif Config.get("otrs_enabled", "OTRS") and not Config.get("yes"):
            menu = Menu(title="What sending method do we want to use?", callbacks=False, fullscreen=True)
            menu.add("Send by SMTP...")
            menu.add("Send by OTRS...")
            o = menu.sout()
            clear()
            if o == '1':
                method = "smtp"
            elif o == '2':
                method = "otrs"
            else:
                print("Unknown option")
                return

        if method == "otrs":
            sender = MailSenderOtrs(self.parser)
            sender.assure_tokens()
            self.wrapper.save()
        elif method == "smtp":
            sender = MailSenderSmtp(self.parser)
        else:
            raise RuntimeError("Unknown sending method: {method}")

        # sending dialog loop
        st = self.parser.stats
        limit = float("inf")
        def limitable(max_): return f"limited to: {limit}/{max_}" if limit < max_ else max_

        while True:
            # clear screen
            info = []

            # display info
            def display_recipients(abroad, text):
                draft = "abroad" if abroad else "local"
                c = st[draft]
                if sum(c) > 0:
                    info.append(f"{text}")
                    if c[0]:
                        info.append(f"Recipient list ({c[0]}/{sum(c)}): "
                                    + ", ".join([o.mail for o in Attachment.get_all(abroad, False, 5, True)]))
                    if c[1]:
                        info.append(f"Already sent ({c[1]}/{sum(c)}): "
                                    + ", ".join([o.mail for o in Attachment.get_all(abroad, True, 5, True)]))
                    info.append(f"Attachment: " + (", ".join(filter(None, (Config.get('attach_files', 'SMTP', get=bool) and "split CSV file attached",
                                                                           Config.get('attach_paths_from_path_column', 'SMTP', get=bool) and "files from the path column attached")))
                                                   or "nothing attached"))
                    info.append(f"\n{Contacts.mail_draft[draft].get_mail_preview()}\n")
                    return True
                return False

            seen_local = display_recipients(False, "  *** E-mail template ***")
            seen_abroad = display_recipients(True, "  *** Abroad template ***")

            if Config.is_testing():
                info.append(
                    f"\n\n\n*** TESTING MOD - mails will be sent to the address: {Config.get('testing_mail')} ***"
                    f"\n (For turning off testing mode set `testing = False` in config.ini.)")
            info.append("*" * 50)
            sum_ = st['local'][0] + st['abroad'][0]
            everything_sent = False
            if sum_ < 1:
                everything_sent = True
                info.append("No e-mail to be sent.")
            method_s = "OTRS" if method == "otrs" else Config.get("smtp_host", "SMTP")

            if test_attachment:
                option = "test"
            elif send_now:  # XConfig.get("yes")
                option = "1"
                send_now = False  # while-loop must not re-send
            else:
                clear()
                menu = Menu("\n".join(info), callbacks=False, fullscreen=False, skippable=False)

                if seen_local or seen_abroad:
                    menu.add(f"Send all e-mails ({limitable(sum_)}) via {method_s}", key="1")
                if seen_local and seen_abroad:
                    menu.add(f"Send local e-mails ({limitable(st['local'][0])})", key="2")
                    menu.add(f"Send abroad e-mails ({limitable(st['abroad'][0])})", key="3")
                if len(menu.menu) == 0:
                    print("No e-mails in the set. Cannot send.")

                t = f" from {limit}" if limit < float("inf") else ""
                menu.add(f"Limit sending amount{t} to...", key="l")
                menu.add("Edit template...", key="e")
                menu.add("Choose recipients...", key="r")
                menu.add("Send test e-mail...", key="t", default=not everything_sent)
                menu.add("Print e-mails to a file...", key="p")
                menu.add(f"Attach files (toggle): {Config.get('attach_files', 'SMTP', get=bool)}", key="a")
                menu.add("Attach paths from path column (toggle):"
                         f" {Config.get('attach_paths_from_path_column', 'SMTP', get=bool)}", key="i")
                menu.add("Go back...", key="x", default=everything_sent)

                option = menu.sout()
                if option is None:
                    return

            # sending menu processing - all, abuse and abroad e-mails
            limit_csirtmails_when_all = limit - st['local'][0]
            if option in ("1", "2") and st['local'][0] > 0:
                print("Sending e-mails...")
                sender.send_list(Attachment.get_all(False, False, limit))
            if option in ("1", "3") and st['abroad'][0] > 0:
                # decrease the limit by the number of e-mails that was just send in basic list
                l = {"1": limit_csirtmails_when_all, "3": limit}[option]
                if l < 1:
                    print("Cannot send abroad e-mails due to limit.")
                else:
                    print("Sending to abroad e-mails...")
                    sender.send_list(Attachment.get_all(True, False, l))
            if option in ["1", "2", "3"]:
                if Config.get("yes"):
                    return
                self.wrapper.save()  # save what message has been sent right now
                input("\n\nPress Enter to continue...")
                continue

            # other menu options
            if option == "e":
                local, abroad = st['local'][0], st['abroad'][0]
                if local:
                    Contacts.mail_draft["local"].edit_text()
                if abroad:
                    Contacts.mail_draft["abroad"].edit_text()
                if not (local or abroad):
                    print("Neither local nor abroad e-mails are to be sent, no editor was opened.")
            elif option == "l":
                limit = ask_number("How many e-mails should be send at once: ")
                if limit < 0:
                    limit = float("inf")
            elif option == "a":
                Config.set("attach_files", not Config.get('attach_files', 'SMTP', get=bool))
            elif option == "i":
                Config.set("attach_paths_from_path_column",
                           not Config.get('attach_paths_from_path_column', 'SMTP', get=bool))
            elif option in ["test", "t", "r", "p"]:
                attachments = sorted(list(Attachment.get_all()), key=lambda x: x.mail.lower())
                if option == "p":
                    with NamedTemporaryFile(mode="w+") as f:
                        try:
                            print(
                                f"The messages are being temporarily generated to the file (stop by Ctrl+C): {f.name}")
                            for attachment in attachments:
                                print(".", end="")
                                sys.stdout.flush()
                                f.write(str(attachment.get_envelope()))
                            f.file.flush()
                            print("Done!")
                        except KeyboardInterrupt:
                            print("Interrupted!")
                        finally:
                            edit(Path(f.name), blocking=True)
                elif option == "test":
                    choices = [o for o in attachments if o.mail == test_attachment]
                    if len(choices) != 1:
                        print(f"Invalid testing attachment {test_attachment}")
                    else:
                        print(choices[0].get_envelope().preview())
                    return
                elif option == "t":
                    # Choose an attachment
                    choices = [(o.mail, "") for o in attachments]
                    try:
                        attachment = attachments[pick_option(choices, f"What attachment should serve as a test?")]
                    except Cancelled:
                        continue
                    clear()

                    # Display generated attachment
                    print(attachment.get_envelope().preview())

                    # Define testing e-mail
                    t = Config.get("testing_mail")
                    t = f" – type in or hit Enter to use {t}" if t else ""
                    try:
                        t = ask(
                            Fore.YELLOW + f"Testing e-mail address to be sent to{t} (Ctrl+C to go back): " + Fore.RESET).strip()
                    except KeyboardInterrupt:
                        continue
                    if not t:
                        t = Config.get("testing_mail")
                    if not t:
                        input("No address written. Hit Enter...")
                        continue

                    # Send testing e-mail
                    old_testing, old_sent, attachment.sent = Config.get('testing'), attachment.sent, None
                    Config.set('testing', True)
                    Config.set('testing_mail', t)
                    sender.send_list([attachment])
                    Config.set('testing', old_testing)
                    input("\n\nTesting completed. Press Enter to continue...")
                    attachment.sent = old_sent
                elif option == "r":
                    # choose recipient list
                    choices = [(o.mail,
                                o.get_draft_name() + ("" if validate_email(o.mail, check_dns=False, check_smtp=False) else " (invalid)"),
                                not o.sent)
                               for o in attachments]
                    code, tags = Dialog(autowidgetsize=True).checklist("Toggle e-mails to be send", choices=choices)
                    if code != 'ok':
                        continue
                    for attachment in attachments:
                        # XX if the same address is going to receive both local and abroad template e-mail,
                        #   this will change the status of its both e-mails.
                        #   In the future, we'd like to have another checklist engine where item references can be passed directly.
                        attachment.sent = attachment.mail not in tags
                    print("Changed!")
            elif option == "x":
                return

    def process(self):
        self.parser.run_analysis()
        self.wrapper.save()

    def config_menu(self):
        def start_debugger():
            raise Debugged

        menu = Menu(title="Config menu")
        menu.add("Edit configuration", lambda: edit("config", 3, restart_when_done=True, blocking=True))
        menu.add("Edit default e-mail template", lambda: edit("template", blocking=True))
        menu.add("Edit default abroad e-mail template", lambda: edit("template_abroad", blocking=True))
        menu.add("Edit uwsgi configuration", lambda: edit("uwsgi"))
        menu.add("Fetch whois for an IP", self.debug_ip)
        menu.add("Start debugger", start_debugger)
        menu.sout()

    def debug_ip(self):
        ip = input("Debugging whois – get IP: ")
        if ip:
            from .whois import Whois
            self.parser.reset_whois(assure_init=True)
            whois = Whois(ip.strip())
            print(whois.analyze())
            print(whois.whois_response)
        input()

    def choose_settings(self):
        """ Remove some of the processing settings """
        actions = []  # description
        discard = []  # lambda to remove the setting
        st = self.parser.settings
        fields = self.parser.fields

        def add_list(labels):
            actions.extend(labels)
            discard.extend((st[type_].pop, i) for i, _ in enumerate(items))

        # Build processing settings list
        for type_, items in st.items():
            if not items and items != 0:
                continue
            if type_ == "split":
                actions.append(f"split by {fields[items]}")
                discard.append((st.pop, "split"))
            elif type_ == "add":
                add_list(f"add {f} (from {str(f.source_field)})" for f in items)
            elif type_ == "filter":
                add_list(f"filter {fields[f].name} {'' if include else '!'}= {val}" for include, f, val in items)
            elif type_ == "unique":
                add_list(f"unique {fields[f].name}" for f in items)
            elif type_ == "aggregate":
                actions.extend(f"{fn.__name__}({fields[col].name})" for fn, col in items[1])
                discard.extend((st[type_][1].pop, i) for i, _ in enumerate(items[1]))
                self.parser.aggregation.clear()  # remove possible aggregation results

        if not actions:
            input("No processing settings found. Hit Enter...")
            return

        # Build dialog
        choices = [(str(i + 1), v, False) for i, v in enumerate(actions)]
        ret, values = Dialog(autowidgetsize=True).checklist("What processing settings should be discarded?",
                                                            choices=choices)
        if ret == "ok":
            # these processing settings should be removed
            for v in values[
                     ::-1]:  # we reverse the list, we need to pop bigger indices first without shifting lower indices
                fn, v = discard[int(v) - 1]
                fn(v)
            if st["aggregate"] and not st["aggregate"][1]:
                # when removing an aggregation settings, we check if it was the last one to get rid of the setting altogether
                del st["aggregate"]

    def redo_menu(self):
        menu = Menu(title="What should be reprocessed?", fullscreen=True)
        menu.add("Delete some processing settings", self.choose_settings)
        menu.add("Delete all processing settings", self.parser.reset_settings)
        menu.add("Delete whois cache", self.parser.reset_whois)
        # XX note that aggregation generators were deleted (jsonpickling) → aggregation count will reset when re-resolving
        # See comment in the Aggregation class, concerning generator serialization.
        menu.add("Resolve unknown abuse-mails", self.parser.resolve_unknown)
        menu.add("Resolve invalid lines", self.parser.resolve_invalid)
        menu.add("Resolve queued lines", self.parser.resolve_queued)
        menu.add("Rework whole file again", self.wrapper.clear)
        menu.sout()



    def close(self):
        self.wrapper.save(last_chance=True)  # re-save cache file
        if not Config.get("yes"):
            if not Config.is_quiet():
                # Build processing settings list
                l = []
                st = self.parser.settings
                fields = self.parser.fields

                for type_, items in st.items():
                    # XX code does not return its custom part
                    if not items and items != 0:
                        continue
                    if type_ == "split":
                        l.append(f"--split {fields[items]}")
                    elif type_ == "add":
                        l.extend(f"--field {f},{str(f.source_field)}" for f in items)
                    elif type_ == "filter":
                        l.extend(f"--{'include' if include else 'exclude'}-filter {fields[f].name},{val}"
                                 for include, f, val in items)
                    elif type_ == "unique":
                        l.extend(f"--unique {fields[f].name}" for f in items)
                    elif type_ == "aggregate":
                        # XX does not work well - at least, they are printed out opposite way
                        l.append(f"--aggregate {items.group_by}," + ",".join(
                            f"{fn.__name__},{col.name}" for fn, col in items.actions))
                if l:
                    print(f" Settings cached:\n convey {self.parser.source_file} " + " ".join(l) + "\n")

            print("Finished.")
        exit(0)

    def cleanup(self):
        """ Make `Controller.run()` calls independent.
        This method is called by ex: tests.
        """
        new_fields.clear()
        Config.cache.clear()

    def get_autocompletion(self, parser):
        actions = [x for action in parser._actions for x in action.option_strings]

        a = ["#!/usr/bin/env bash",
             "# bash completion for convey",
             ""
             '_convey()',
             '{',
             '  local cur',
             '  local cmd',
             '',
             '  cur=${COMP_WORDS[$COMP_CWORD]}',
             '  prev="${COMP_WORDS[COMP_CWORD-1]}";',
             '  cmd=( ${COMP_WORDS[@]} )',
             '',
             '  if [[ "$prev" == -f ]] || [[ "$prev" == --field ]] ||' +
             '  [[ "$prev" == -fe ]] || [[ "$prev" == --field-excluded ]]; then',
             f'        COMPREPLY=( $( compgen -W "{" ".join(t.name for t in Types.get_computable_types())}"  -- "$cur" ) )',
             '        return 0',
             '    fi',
             '',
             '  if [[ "$prev" == -a ]] || [[ "$prev" == --aggregate ]]; then',
             '    param=(${cur//,/ })',
             f'        COMPREPLY=( $( compgen -W "{" ".join("${param[0]}," + s for s in aggregate_functions)}"  -- "$cur" ) )',
             '        return 0',
             '    fi',
             '',
             '  if [[ "$cur" == -* ]]; then',
             f'    COMPREPLY=( $( compgen -W "{" ".join(actions)}" -- $cur ) )',
             '    return 0',
             '  fi',
             '}',
             '',
             'complete -F _convey -o default convey']
        return "\n".join(a)



from __future__ import annotations
from csv import reader as csvreader
from pathlib import Path
import stat
from types import SimpleNamespace
from typing import TYPE_CHECKING

from validate_email import validate_email

from .config import Config
from .contacts import Contacts
from .types import Types

if TYPE_CHECKING:
    from .parser import Parser

class Attachment:
    sent: bool
    "True => sent, False => error while sending, None => not yet sent"

    abroad: bool
    "True => abroad e-mail is in Contacts dict, False => e-mail is file name, None => undeliverable (no e-mail)"

    def __init__(self, filename):
        self.abroad = Config.ABROAD_MARK in filename
        self._sent = None
        self.filename = filename
        self.used_in_body : bool = False
        "Whether the file contents should be added as an e-mail attachment. If set, we suppose, the attachment has been pasted to the body text via jinja template."

        st = self.parser.stats

        if validate_email(self.mail, check_dns=False, check_smtp=False):
            st[self.get_draft_name()][int(bool(self.sent))] += 1
        else:
            st["non_deliverable"] += 1
        st["totals"] += 1

    def get_paths_from_path_column(self):
        """ Load a column that was marked being of the path Type. Checks the file meet some security standards
            that are quite limiting (and possibly be reduced in the future)
            so that a crafted CSV would not pull up ~/.ssh or /etc/ files.

            :raises RuntimeError if there is security concern about an attachment
        """
        path = next((f for f in self.parser.fields if f.type == Types.path), None)

        with self.path.open() as f:
            reader = csvreader(f, dialect=self.parser.settings["dialect"])
            paths = [Path(row[path.col_i]) for row in reader]
            for path in paths:
                if str(path) != path.name:
                    raise RuntimeError(f"For security reasons, path must be just a file name: {path}")
                if not path.stat().st_mode & stat.S_IROTH:
                    raise RuntimeError(f"For security reasons, path must be readable to others: {path}")
                if path.is_symlink():
                    raise RuntimeError(f"For security reasons, path must not be a symlink: {path}")
            return paths

    @property
    def path(self):
        """ Path of the split CSV file (that contain rows having the same value in the split column). """
        return Path(Config.get_cache_dir(), self.filename)

    @property
    def sent(self):
        return self._sent

    @sent.setter
    def sent(self, value):
        if self._sent != value:
            # no effect if sent status changes from None (not tried) to False (not succeeded)
            if bool(self._sent) != bool(value):
                r = self.parser.stats[self.get_draft_name()]
                # exchange the count from the not-sent to sent or vice verse
                r[int(value is not True)] -= 1
                r[int(value is True)] += 1
            self._sent = value

    @classmethod
    def init(cls, parser):
        cls.parser : Parser = parser

    @classmethod
    def reset(cls, stats):
        stats["abroad"] = [0, 0]  # abroad [not-sent, sent]
        stats["local"] = [0, 0]  # local [not-sent, sent]
        stats["non_deliverable"] = 0
        stats["totals"] = 0

    @classmethod
    def get_all(cls, abroad=None, sent=None, limit=float("inf"), threedots=False):
        for i, o in enumerate(cls.parser.attachments):
            # we want to filter either sent or not-sent attachments only
            if sent is not None and sent is not bool(o.sent):
                continue
            if abroad is not None and any((abroad and not o.abroad, not abroad and o.abroad)):
                # filtering abroad/local attachments only
                continue
            if limit == 0:
                if threedots:
                    yield SimpleNamespace(mail="...")
                return
            else:
                limit -= 1
            yield o

    @property
    def mail(self):
        if self.abroad:
            return self.filename[self.filename.index(Config.ABROAD_MARK) + 2:]
        return self.filename

    @property
    def cc(self):
        """ Return cc header from Contacts.
            This has nothing in common with possible Cc header in the mail_draft template!
        """
        return Types.get_method(Types.abusemail, Types.cc_contact)(self.mail)

    def get_draft_name(self):
        return "abroad" if self.abroad else "local"

    def get_draft(self):
        return Contacts.mail_draft[self.get_draft_name()]

    def get_envelope(self):
        return self.get_draft().get_envelope(self)



import bdb
import importlib.util
import ipaddress
import logging
import re
import socket
import subprocess
from abc import ABCMeta
from base64 import b64decode, b64encode
from builtins import ZeroDivisionError
from copy import copy
from datetime import datetime
from enum import IntEnum
from pathlib import Path
from quopri import decodestring, encodestring
from sys import exit
from typing import List
from urllib.parse import unquote, quote, urlparse, urlsplit, urljoin

import dateutil.parser
import requests
import urllib3
from bs4 import BeautifulSoup
from chardet import detect
from netaddr import IPRange, IPNetwork
from pint import UnitRegistry
from validate_email import validate_email

from .config import Config
from .contacts import Contacts
from .decorators import PickBase, PickMethod, PickInput
from .graph import Graph
from .infodicts import phone_regex_match, phone_country, address_country, country_codes
from .utils import timeout, print_atomic
from .whois import Whois

logger = logging.getLogger(__name__)

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)  # seen due to web module requests.get(verify=False)

reIpWithPort = re.compile("((\d{1,3}\.){4})(\d+)")
reAnyIp = re.compile("\"?((\d{1,3}\.){3}(\d{1,3}))")
reFqdn = re.compile(
    "(?=^.{4,253}$)(^((?!-)[a-zA-Z0-9-_]{1,63}(?<!-)\.)+[a-zA-Z]{2,63}$)")  # Xtoo long, infinite loop: ^(((([A-Za-z0-9]+){1,63}\.)|(([A-Za-z0-9]+(\-)+[A-Za-z0-9]+){1,63}\.))+){1,255}$
reUrl = re.compile('[htps]*://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+')

# reBase64 = re.compile('^([A-Za-z0-9+/]{4})*([A-Za-z0-9+/]{4}|[A-Za-z0-9+/]{3}=|[A-Za-z0-9+/]{2}==)$')


types: List["Type"] = []  # all field types
methods = {}
graph = Graph()


class Aggregate:
    @staticmethod
    def avg():
        count = 0
        res = float((yield))
        while True:
            count += 1
            res += float((yield res / count))

    @staticmethod
    def sum():
        res = 0
        while True:
            res += float((yield res))

    @staticmethod
    def count():
        res = 0
        while True:
            yield res
            res += 1

    @staticmethod
    def min():
        v = yield
        while True:
            v = min((yield v), v)

    @staticmethod
    def max():
        v = yield
        while True:
            v = max((yield v), v)

    @staticmethod
    def list():
        l = []
        while True:
            l.append((yield l))

    @staticmethod
    def set():
        s = set()
        while True:
            s.add((yield s))

    # XX If we would like to serialize a function and this is not possible, we can serialize it ourselves that way:
    # @staticmethod
    # def avg():
    #     count = 0
    #     try:
    #         res = float((yield))
    #     except LoadFromSerialization:
    #         count, res = yield
    #     try:
    #         while True:
    #             count += 1
    #             res += float((yield res / count))
    #     except StopIteration: -> serialization request
    #         yield count, res


def is_ip(ip):
    """ True, if IP is well formatted IPv4 or IPv6 """
    try:
        ipaddress.ip_address(ip)
        return True
    except ValueError:
        return False


pint = UnitRegistry()


class Checker:
    """ To not pollute the namespace, we put the methods here """

    hostname_ips_cache = {}
    hostname_cache = {}

    @staticmethod
    def hostname_ips(val):
        if val not in Checker.hostname_ips_cache:
            try:
                Checker.hostname_ips_cache[val] = list(
                    {addr[4][0] for addr in timeout(15, socket.getaddrinfo, val, None)})
            except (TimeoutError, OSError) as e:
                Checker.hostname_ips_cache[val] = []
        return Checker.hostname_ips_cache[val]

    @classmethod
    def hostname_ip(cls, val):
        if val not in cls.hostname_cache:
            try:
                cls.hostname_cache[val] = timeout(3, socket.gethostbyname, val)
            except (TimeoutError, OSError) as e:
                logger.warning(f"Hostname {val}: {e}")
                cls.hostname_cache[val] = []
        return cls.hostname_cache[val]

    @staticmethod
    def check_cidr(cidr):
        try:
            ipaddress.ip_interface(cidr)
            try:
                ipaddress.ip_address(cidr)
            except ValueError:  # "1.2.3.4" fail, "1.2.3.4/24" pass
                return True
        except ValueError:
            return False

    @staticmethod
    def is_base64(x):
        """ We prefer as base64-encoded only such strings that could be decoded to UTF-8.
            Because otherwise nearly any ASCII input with the correct padding
            would be considered as base64 (ex: port number), even when well readable at first sight.
            If not UTF-8 decodable, we check the minimal length and penalize.
        """
        if not re.match(r"^(?:[A-Za-z0-9+/]{4})*(?:[A-Za-z0-9+/]{2}==|[A-Za-z0-9+/]{3}=)?$", x):
            return False

        try:
            s = b64decode(x).decode("UTF-8")
            if s:
                if len(re.sub(r"[^a-zA-Z0-9 ]", "", s)) / len(s) > 0.9:
                    # Majority of characters in the decoded string must be alphanumerical
                    # in order we think this might be a base64-encoded string.
                    # Imagine string 'ahojahoj' which is base64-decodable but gibberish.
                    return True
        except (UnicodeDecodeError, ValueError):
            try:
                if len(x) > 10 and b64decode(x):
                    # if the string is long enough, we admit it can be encoded in another charset
                    return -0.5
            except (UnicodeDecodeError, ValueError):
                pass

    @staticmethod
    def is_quopri(x):
        try:
            # When 'rmx1u2916Gv9IGv58iBw7Gwg7+FiZWxza+kg82R5Lg==' is decoded, only last '=' is stripped out.
            decoded = decodestring(x).decode("UTF-8").rstrip("=")
            return decoded and decoded != x.rstrip("=")
        except (UnicodeDecodeError, ValueError):
            return False

    @staticmethod
    def bytes_plaintext(x):
        try:
            return x.decode("UTF-8").replace("\n", r"\n")
        except UnicodeDecodeError:
            enc = detect(x)["encoding"]
            # Czech phrase: "Žluťoučký kůň pěl ďábelské ódy.".encode("iso-8859-2")
            # results in turkish {'encoding': 'ISO-8859-9', 'confidence': 0.47567063613812527, 'language': 'Turkish'}
            # which is not perfect but better than nothing.
            if enc:
                return x.decode(enc)
            return x

    @staticmethod
    def is_urlencode(x):
        return unquote(x) != x

    @staticmethod
    def check_wrong_url(wrong):
        if (reUrl.match(wrong) and "[.]" not in wrong) or reFqdn.match(wrong) or is_ip(wrong):
            # input "example[.]com" would be admitted as a valid URL)
            return False
        s = wrong_url_2_url(wrong, make=False)
        s2 = wrong_url_2_url(wrong, make=True)
        try:
            netloc = urlparse(s2).netloc
        except ValueError:
            # ex: s2='http://[07/Feb/2021:00:32:43 -0500] "GET /wp-json/wp/v2/posts HTTP/1.1" 200 1912'
            return False
        # Xm = reUrl.match(s2)
        if reFqdn.match(s) or reFqdn.match(netloc) or is_ip(netloc):  # X(m and m.span(0)[1] == len(s2) and "." in s2):
            # Xwe impose full match
            # Xstring created from "x"*100 would be submitted for a valid URL
            return True
        return False

    @staticmethod
    @PickInput
    def decode(val, encoding="utf-8"):
        if type(val) is str:
            val = val.encode("utf-8")
        return val.decode(encoding).replace("\n", r"\n")

    @staticmethod
    @PickInput
    def time_format(val, fmt):
        return dateutil.parser.parse(val, fuzzy=True).strftime(fmt)

    # noinspection PyBroadException
    @staticmethod
    def is_unit(val):
        if len(val) == 1:
            # we want prevent "a" being considered a time unit ("a" == "1 year" according to pint 0.16.1)
            return False
        try:
            e = pint.parse_expression(val)
            return e and pint.get_compatible_units(e)
        except Exception:
            return False

    @staticmethod
    @PickInput
    def unit_expand(val, unit=None):
        try:
            e = pint.parse_expression(val)
            if unit:
                return str(e.to(unit))
            else:
                l = []
                for unit in pint.get_compatible_units(e):
                    l.append(str(e.to(unit)))
                return l
        except (ValueError, AttributeError):
            return ""

    @classmethod
    def is_phone(cls, val):
        return phone_regex_match(val) and not cls.is_timestamp(val)

    @staticmethod
    def is_timestamp(val):
        reference = datetime.now()

        # identify UNIX time
        try:
            o = datetime.fromtimestamp(float(val))
        except ValueError:
            pass  # val is not a number, try different methods
        else:
            # consider only nowadays timestamp a timestamp
            # prevent "12345" to be considered as a timestamp from the beginning of the Unix epoch (1970)
            return reference.year - 5 < o.year < reference.year + 1

        # identify straight textual representation (ex: 2021-02-12), easily parsable by not-fuzzy `dateutil`
        try:
            o = dateutil.parser.parse(val)
            if not (o.second == o.minute == o.hour == 0 and o.day == reference.day and o.month == reference.month):
                # if parser accepts a plain number, ex: 1920,
                # it thinks this is a year without time (assigns midnight) and without date (assigns current date)
                # -> 1920-05-05 00:00 (on 5th May) – we try to skip such result.
                # If does not match, it seems this is a valid timestamp and we hereby return True.
                return True
        except ValueError:
            pass  # val is a random textual representation or a UNIX time
        except OverflowError:
            return False

        # identify random textual representation of a timestamp
        try:
            o = Checker.parse_timestamp(val)
            if 2100 > o.year > 1900 and not (o.second == o.minute == o.hour == 0 and o.day == o.month == 1):
                # this year seems reasonable and it is not suspicious
                # fuzzy search often return crazy things like year from port number - these records have 1 Jan midnight
                return True
        except (ValueError, TypeError):
            pass
        return False

    default_datetime = datetime(1, 1, 1)

    @staticmethod
    def parse_timestamp(val):
        # UNIX time
        try:
            return datetime.fromtimestamp(float(val))
        except ValueError:
            pass  # val is not a number

        # random textual representation of a timestamp
        try:
            return dateutil.parser.parse(val, fuzzy=True, default=Checker.default_datetime)
        except OverflowError:
            return Checker.default_datetime

    @staticmethod
    def isotimestamp(val):
        o = Checker.parse_timestamp(val)
        s = o.isoformat()
        if o.year == 1:  # this is a fake year from our default_datetime object
            s = "0000-00-00" + s[10:]
        return s

    @staticmethod
    def date(val):
        o = Checker.parse_timestamp(val)
        if o.year == 1:  # this is a fake year from our default_datetime object
            return ""
        return o.date()

    @staticmethod
    @PickMethod("all")
    class HostnameTld:
        @staticmethod
        def all(x):
            """ take all TLD """
            x = x[x.rindex(".") + 1:]
            if not x.isdigit():
                return x

        @classmethod
        def ccTLD(cls, x):
            """ country code only """
            x = cls.all(x)
            return x if len(x) == 2 else ""

        @classmethod
        def gTLD(cls, x):
            """ generic only """
            x = cls.all(x)
            return x if len(x) != 2 else ""

    @staticmethod
    def prefix_cidr(val):
        if "/" in val:
            # currently, we do not guarantee prefix will be in the form "... - ...",
            # it may has directly CIDR form if WHOIS tells us so
            return val
        return IPRange(*val.split("-")).cidrs()

    @staticmethod
    def cidr_ips(val):
        return [ip for ip in IPNetwork(val)]


def wrong_url_2_url(s, make=True):
    s = re.sub("^hxxp", 'http', s, flags=re.I).replace("[.]", ".").replace("(.)", ".").replace("[:]", ":")
    if make and not s.lower().startswith("http"):
        s = "http://" + s
    return s


def any_ip_ip(s):
    m = reAnyIp.search(s)
    if m:
        return m.group(1)


def port_ip_ip(s):
    m = reIpWithPort.match(s)
    if m:
        return m.group(1).rstrip(".")


def port_ip_port(s):
    m = reIpWithPort.match(s)
    if m:
        return m.group(3)


def url_port(s):
    s = s.split(":")[1]
    return re.match("^(\d*)", s).group(1)


class Web:
    """
    :return: self.get = [http status | error, shortened text, original html, redirects, x-frame-options, csp, form_names]
    """
    cache = {}
    store_html = True
    store_text = True
    headers = {}

    @classmethod
    def init(cls, used_types: List = None):
        if used_types:
            cls.store_html = Types.html in used_types
            cls.store_text = Types.text in used_types
        else:
            cls.store_html = cls.store_text = True
        if Config.get("user_agent", "FIELDS"):
            cls.headers = {"User-Agent": Config.get("user_agent", "FIELDS")}

    def __init__(self, url):
        if url in self.cache:
            self.get = self.cache[url]
            return
        redirects = []
        current_url = url
        while True:
            try:
                logger.debug(f"Scrapping connection to {current_url}")
                response = requests.get(current_url, timeout=Config.get("web_timeout", "FIELDS", get=int),
                                        headers=self.headers,
                                        allow_redirects=False, verify=False)
            except IOError as e:
                if isinstance(e, requests.exceptions.HTTPError):
                    s = 0
                elif isinstance(e, requests.exceptions.ConnectionError):
                    s = -1
                elif isinstance(e, requests.exceptions.RequestException):
                    s = -2
                elif isinstance(e, requests.exceptions.Timeout):
                    s = -3
                else:
                    s = e
                self.cache[url] = self.get = str(s), None, None, redirects, None, None, None
                print_atomic(f"Scrapping {url} failed: {e}")
                break
            if response.headers.get("Location") and len(redirects) < 10:
                current_url = urljoin(current_url, response.headers.get("Location"))
                redirects.append(current_url)
                continue
            else:
                response.encoding = response.apparent_encoding  # https://stackoverflow.com/a/52615216/2036148
                if self.store_text:
                    soup = BeautifulSoup(response.text, features="html.parser")
                    # check redirect
                    res = soup.select("meta[http-equiv=refresh i]")
                    if res:
                        wait, txt = res[0].attrs["content"].split(";")
                        m = re.search(r"http[^\"'\s]*", txt)
                        if m and len(redirects) < 10:
                            current_url = m.group(0)
                            redirects.append(current_url)
                            continue
                    # prepare content to be shortened
                    [s.extract() for s in
                     soup(["style", "script", "head"])]  # remove tags with low probability of content
                    text = re.sub(r'\n\s*\n', '\n', soup.text)  # reduce multiple new lines to singles
                    text = re.sub(r'[^\S\r\n][^\S\r\n]*[^\S\r\n]', ' ',
                                  text)  # reduce multiple spaces (not new lines) to singles

                    # if the form tag like <input> or <select> has no attribute "name", print out its tag name and value or options
                    def get_info(el):
                        """ This element has no "name" attribute """
                        n = el.name
                        r = [n]
                        if n == "select":
                            for opt in el.find_all("option"):
                                r.append(opt.attrs.get("value", "") + ":" + opt.text)
                        else:
                            r.append(el.attrs.get("value", ""))
                        return " ".join(r)

                    form_names = [s.attrs.get("name", get_info(s)) for s in soup(("input", "select", "textarea"))]
                else:
                    form_names = None
                    text = ""
                # for res in response.history[1:]:
                #     redirects += f"REDIRECT {res.status_code} → {res.url}\n" + text
                #     redirects.append(res.url)

                print_atomic(f"Scrapped {url} ({len(response.text)} bytes)")
                self.cache[
                    url] = self.get = response.status_code, text.strip(), response.text if self.store_html else None, \
                                      redirects, \
                                      response.headers.get('X-Frame-Options', None), \
                                      response.headers.get('Content-Security-Policy', None), \
                                      form_names
                break
                # if current_url == url:
                #     break
                # url = current_url


def dig(rr):
    def dig_query(query):
        print_atomic(f"Digging {rr} of {query}")
        if rr == "SPF":
            t = "TXT"
        elif rr == "DMARC":
            query = "_dmarc." + query
            t = "TXT"
        else:
            t = rr
        try:
            text = subprocess.check_output(["dig", "+short", "-t", t, query, "+timeout=1"]).decode("utf-8")
        except FileNotFoundError:
            Config.missing_dependency("dnsutils")
        if text.startswith(";;"):
            return None
        spl = text.split("\n")[:-1]
        if t == "TXT":
            spl = [r[1:-1] for r in spl if r.startswith('"') and r.endswith('"')]  # row without " may be CNAME redirect
            if rr == "SPF":
                return [r for r in spl if r.startswith('v=spf')]
            elif rr == "TXT":
                return [r for r in spl if not r.startswith('v=spf')]
        logger.debug(f"Dug {spl}")
        return spl

    return dig_query


# @PickInput XX not yet possible to PickInput
# 1. since for port '80' wizzard loads the port '8' and then '80'
# 2. since we cannot specify `--field ports[80,443]` because comma ',' is taken for a delmiiter between FIELD and COLUMN
#   and pair quoting char '[]' is not allowed in csv.reader that parses CLI
def nmap(val, port=""):
    """
    :type port: int Port to scan, you may delimit by a comma. Ex: `80, 443`
    """
    logger.info(f"NMAPing {val}...")
    try:
        cmd = ["nmap", val]
        if port:
            cmd.extend(["-p", port])
        text = subprocess.run(cmd, stdout=subprocess.PIPE).stdout.decode("utf-8")
    except FileNotFoundError:
        Config.missing_dependency("nmap")
    text = text[text.find("PORT"):]
    text = text[text.find("\n") + 1:]
    text = text[:text.find("\n\n")]
    if Config.get("multiple_nmap_ports", "FIELDS"):
        l = []
        for row in text.split("\n"):
            l.append(int(re.match("(\d+)", row).group(1)))
        return l

    return text


methods_deleted = {}


class TypeGroup(IntEnum):
    general = 1
    custom = 2
    whois = 3
    dns = 4
    nmap = 5
    web = 6

    @staticmethod
    def init():
        for module in ["whois", "web", "nmap", "dig"]:
            if Config.get(module, "FIELDS") is False:
                if module == "dig":
                    module = "dns"
                getattr(TypeGroup, module).disable()

    def disable(self):
        for start, target in copy(methods):
            if start.group is self or target.group is self:
                methods_deleted[start, target] = methods[start, target]
                del methods[start, target]
        for t in Types.get_guessable_types():
            if t.group is self:
                t.is_disabled = True


class Type:
    """
    A field type Convey is able to identify or compute
    """

    def __init__(self, name, group=TypeGroup.general, description="", usual_names=[], identify_method=None,
                 is_private=False,
                 from_message=None, usual_must_match=False):
        """
        :param name: Key name
        :param description: Help text
        :param usual_names: Names this column usually has (ex: source_ip for an IP column). List of str, lowercase, no spaces.
        :param identify_method: Lambda used to identify a value conforms this field type (return true).
            If float is returned, this should be added to the score (ex: malus when we are not sure this is a base64 string).
        :param is_private: User cannot add the field type (ex: whois, user can extend only netname which is accessed through it).
        """
        self.name = name
        self.group = group
        self.usual_names = usual_names
        self.usual_must_match = usual_must_match
        self.identify_method = identify_method
        self.description = description
        self.is_private = is_private
        self.is_disabled = False  # disabled field cannot be added (and computed)
        self.from_message = from_message
        types.append(self)
        # Type considered equal when starting computing.
        # Ex: hostname from source_ip will be computed as if from an ip.
        # So that Types.source_ip will have `computing_start = ip`
        self.computing_start: Type = self
        # Types considered equal to be computed from.
        # Ex: hostname from ip may be computed even from a source_ip.
        # So that Types.ip will have `equals = [ip, source_ip]`
        self.equals: List["Types"] = [self]
        self.is_plaintext_derivable = False

    def __getstate__(self):
        return str(self)

    def __setstate__(self, state):
        self.__dict__.update(getattr(Types, state).__dict__)

    def __eq__(self, other):
        if other is None:
            return False
        if type(other) is str:
            return self.name == other
        return self.name == other.name

    def __lt__(self, other):
        if isinstance(other, str):
            return self.name < other
        return (self.group, self.name) < (other.group, other.name)

    def __hash__(self):
        return hash(self.name)

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"Type({self.name})"

    def doc(self):
        s = self.name
        if self.description:
            s += f" ({self.description})"
        if self.usual_names:
            s += f" usual names: " + ", ".join(self.usual_names)
        return s

    def __add__(self, other):  # sometimes we might get compared to string columns names from the CSV
        if isinstance(other, str):
            return self.name + " " + other
        return self.name + " " + other.name

    def __radd__(self, other):  # sometimes we might get compared to string columns names from the CSV
        if isinstance(other, str):
            return other + " " + self.name
        return self.other + " " + self.name

    def init(self):
        """ Init self.computing_start and self.equals . """
        computing_start = [stop for start, stop in methods if start is self and methods[start, stop] is True]
        equals = [start for start, stop in methods if stop is self and methods[start, stop] is True]
        if len(computing_start) == 1:
            self.computing_start = computing_start[0]
        elif len(computing_start):
            raise ConnectionAbortedError(f"Multiple 'computing_start' types defined for {self}: {computing_start}")
        if equals:
            self.equals = equals + [self]

        # check if this is plaintext derivable
        if self != Types.plaintext:
            self.is_plaintext_derivable = bool(graph.dijkstra(self, start=Types.plaintext))

    def check_conformity(self, samples, has_header, field):
        """
        :rtype: int|False Score if the given info conforms to this type.
        """
        score = 0
        # print("Guess:", key, names, checkFn)
        # guess field type by name

        if has_header:
            s = str(field).replace(" ", "").replace("'", "").replace('"', "").lower()
            for n in self.usual_names:
                if s in n or n in s:
                    # print("HEADER match", field, self, self.usual_names)
                    score += 2 if self.usual_must_match else 1
                    break
        if not score and self.usual_must_match:
            return False
        # else:
        # guess field type by few values
        hits = 0
        for val in samples:
            sc = self.identify_method(val)
            if sc:
                hits += 1
                if isinstance(sc, float):
                    score += sc

        try:
            percent = hits / len(samples)
        except ZeroDivisionError:
            percent = 0
        if percent == 0:
            return False
        elif percent > 0.6:
            # print("Function match", field, checkFn)
            score += 1
            if percent > 0.8:
                score += 1
        return score


class Types:
    """
    Methods sourcing from private type should return a tuple, with the private type as the first
    Ex: (whois, asn): lambda x: (x, x.get[3])

    """

    @staticmethod
    def refresh():
        """ refreshes methods and import custom methods from files """
        methods.clear()
        methods.update(Types._get_methods())
        graph.clear()
        [graph.add_edge(to, from_) for to, from_ in methods if methods[to, from_] is not True]
        [t.init() for t in types]

        if Config.get("disable_external", get=bool) is True:
            return

        try:
            externals = Config.config["EXTERNAL"]
        except KeyError:
            externals = []
        for field_name in externals:
            if field_name == "external_fields":  # this is a genuine field, user did not make it
                continue
            path, method_name = Config.config["EXTERNAL"][field_name].rsplit(":")
            module = get_module_from_path(path)
            Types.import_method(module, method_name, path, name=field_name)

        for path in (x.strip() for x in Config.get("external_fields", "EXTERNAL", get=str).split(",") if x.strip()):
            # noinspection PyBroadException
            try:
                module = get_module_from_path(path)
                if module:
                    for method_name in (x for x in dir(module) if not x.startswith("_")):
                        Types.import_method(module, method_name, path)
            except bdb.BdbQuit:
                raise
            except Exception as e:
                s = f"Cannot import custom file from path: {path}"
                input(s + ". Press any key...")
                logger.warning(s)

    @staticmethod
    def import_method(module, method_name, path, name=None):
        if not name:
            name = method_name
        lambda_ = getattr(module, method_name)
        if isinstance(lambda_, ABCMeta):  # this is just an import statement of ex: PickBase
            return
        doc = lambda_.__doc__ if not isinstance(lambda_, PickBase) else lambda_.get_type_description()
        # if Config.get("disable_external", get=bool) is True:
        #     # user do not want to allow externals to be added but we have added them before argparse was parsed
        #     # so that we could inform the user in the help text of the computable fields
        #     types.pop(types.index(getattr(Types, name)))
        #     delattr(Types, name)
        # else:
        setattr(Types, name, Type(name, TypeGroup.general, doc))
        type_ = getattr(Types, name)
        methods[(Types.plaintext, type_)] = lambda_
        if lambda_ is not True:
            graph.add_edge(Types.plaintext, type_)
        type_.init()
        # Xif Config.get("disable_external", get=bool) is False:
        # Xwhen "disable_external" is None it means this method is called before argparse flags are parsed,
        # Xwe do not know yet if "disable_external" will be set to True or False
        logger.debug(f"Successfully added method {method_name} from module {path}")

    @staticmethod
    def find_type(source_type_candidate):
        source_type = None
        source_t = source_type_candidate.lower().replace(" ", "")  # make it seem like a usual field name
        possible = None
        for t in types:
            if source_t == t:  # exact field name
                source_type = t
                break
            if source_t in t.usual_names:  # usual field name
                possible = t
        else:
            if possible:
                source_type = possible
        return source_type

    @staticmethod
    def get_method(start: Type, target: Type):
        try:
            return methods[start, target]
        except KeyError:
            if (start, target) in methods_deleted:
                print(f"Disabled path at {start} – {target}. Launch --config to enable it.")
                exit()
            else:
                raise LookupError

    whois = Type("whois", TypeGroup.whois, "ask whois servers", is_private=True)
    web = Type("web", TypeGroup.web, "scrape web contents", is_private=True)

    external = Type("external", TypeGroup.custom, from_message="from a method in your .py file")
    code = Type("code", TypeGroup.custom, from_message="from a code you write")
    reg = Type("reg", TypeGroup.custom, from_message="from a regular expression")
    reg_s = Type("reg_s", TypeGroup.custom, from_message="substitution from a regular expression")
    reg_m = Type("reg_m", TypeGroup.custom, from_message="match from a regular expression")
    netname = Type("netname", TypeGroup.whois)
    country = Type("country", TypeGroup.whois)
    abusemail = Type("abusemail", TypeGroup.whois, "Abuse e-mail contact from whois")
    prefix = Type("prefix", TypeGroup.whois)  # XX rename to 'inetnum'? to 'range'?
    csirt_contact = Type("csirt_contact", TypeGroup.whois,
                         "E-mail address corresponding with country code, taken from your personal contacts_abroad CSV"
                         " in the format `country,abusemail`. See config.ini/contacts_abroad")
    incident_contact = Type("incident_contact", TypeGroup.whois)
    cc_contact = Type("cc_contact", TypeGroup.whois,
                      "E-mail address corresponding with the abusemail, taken from your personal contacts_cc CSV"
                      " in the format `domain,cc;cc` (mails delimited by a semicolon). See config.ini/contacts_cc."
                      " Note:  when sending an e-mail to a csirt_contact, the e-mail template fills Cc header"
                      " from the csirt_contact through contacts_cc.csv too. So you may end up with another Cc field"
                      " than in cc_contact.")
    text = Type("text", TypeGroup.web)
    http_status = Type("http_status", TypeGroup.web, "HTTP response status. If 0 or negative, request failed.")
    html = Type("html", TypeGroup.web)
    redirects = Type("redirects", TypeGroup.web)
    x_frame_options = Type("x_frame_options", TypeGroup.web)
    csp = Type("csp", TypeGroup.web)
    form_names = Type("form_names", TypeGroup.web)
    ports = Type("ports", TypeGroup.nmap, "Open ports given by nmap")
    spf = Type("spf", TypeGroup.dns)
    txt = Type("txt", TypeGroup.dns)
    a = Type("a", TypeGroup.dns)
    aaaa = Type("aaaa", TypeGroup.dns)
    ns = Type("ns", TypeGroup.dns)
    mx = Type("mx", TypeGroup.dns)
    dmarc = Type("dmarc", TypeGroup.dns)
    tld = Type("tld", TypeGroup.general)
    formatted_time = Type("formatted_time", TypeGroup.general)
    isotimestamp = Type("isotimestamp", TypeGroup.general)
    time = Type("time", TypeGroup.general)
    date = Type("date", TypeGroup.general)
    bytes = Type("bytes", TypeGroup.general, is_private=True)
    charset = Type("charset", TypeGroup.general)
    country_name = Type("country_name",
                        TypeGroup.general)  # XX not identifiable, user has to be told somehow there is such method
    phone = Type("phone", TypeGroup.general, "telephone number", ["telephone", "tel"], Checker.is_phone)

    unit = Type("unit", TypeGroup.general, "any physical quantity", [], Checker.is_unit)
    timestamp = Type("timestamp", TypeGroup.general, "time or date", ["time"], Checker.is_timestamp)
    ip = Type("ip", TypeGroup.general, "valid IP address", ["ip", "ipaddress"], is_ip)
    source_ip = Type("source_ip", TypeGroup.general, "valid source IP address",
                     ["sourceipaddress", "source", "src"], is_ip, usual_must_match=True)
    destination_ip = Type("destination_ip", TypeGroup.general, "valid destination IP address",
                          ["destinationipaddress", "destination", "dest", "dst"], is_ip, usual_must_match=True)
    port = Type("port", TypeGroup.general, "port", ["port", "prt"], lambda x: re.match("\d{1,5}", x),
                usual_must_match=True)
    cidr = Type("cidr", TypeGroup.general, "CIDR 127.0.0.1/32", ["cidr"], Checker.check_cidr)
    port_ip = Type("port_ip", TypeGroup.general, "IP in the form 1.2.3.4.port", [], reIpWithPort.match)
    any_ip = Type("any_ip", TypeGroup.general, "IP in the form 'any text 1.2.3.4 any text'", [],
                  lambda x: reAnyIp.search(x) and not is_ip(x))
    hostname = Type("hostname", TypeGroup.general, "2nd or 3rd domain name", ["fqdn", "hostname", "domain"],
                    reFqdn.match)
    email = Type("email", TypeGroup.general, "E-mail address", ["mail"],
                 lambda e: validate_email(e, check_dns=False, check_smtp=False))
    url = Type("url", TypeGroup.general, "URL starting with http/https", ["url", "uri", "location"],
               lambda s: reUrl.match(s) and "[.]" not in s)  # input "example[.]com" would be admitted as a valid URL)
    asn = Type("asn", TypeGroup.whois, "Autonomous system number", ["as", "asn", "asnumber"],
               lambda x: re.search('AS\d+', x) is not None)
    base64 = Type("base64", TypeGroup.general, "Text encoded with Base64", ["base64"], Checker.is_base64)
    quoted_printable = Type("quoted_printable", TypeGroup.general, "Text encoded as quotedprintable", [],
                            Checker.is_quopri)
    urlencode = Type("urlencode", TypeGroup.general, "Text encoded with urlencode", ["urlencode"], Checker.is_urlencode)
    wrong_url = Type("wrong_url", TypeGroup.general, "Deactivated URL", [], Checker.check_wrong_url)
    plaintext = Type("plaintext", TypeGroup.general, "Plain text", ["plaintext", "text"], lambda x: False)
    path = Type("path", usual_names=["path"], identify_method = lambda x: False)


    @staticmethod
    def get_computable_types(ignore_custom=False):
        """ List of all suitable fields that we may compute from a suitable output
        :type ignore_custom: bool Ignore types that may not be reachable because user input would be needed.
            These are TypeGroup.custom and instances of PickBase.
        """
        s = set()
        for (_, target_type), m in methods.items():
            if ignore_custom and (target_type.group == TypeGroup.custom or (isinstance(m, PickBase) and not m.default)):
                continue
            if target_type.is_private or target_type.is_disabled:
                continue
            s.add(target_type)
        return sorted(s)

    @staticmethod
    def get_guessable_types() -> List[Type]:
        """ these field types can be guessed from a string """
        return sorted([t for t in types if not t.is_disabled and (t.identify_method or t.usual_names)])

    @staticmethod
    def get_uml(flags):
        """ Return DOT UML source code of types and methods
            dashed nodes = identifiable automatically
            dashed edges = lambda-True relation (see _get_methods() help)
            squares = category border
            edge label = possible ways of processing (@PickBase decorated methods)

          FLAGS:
             * +1 to gray out disabled fields/methods
             * +2 to include usual field names
             * +16 wide-screen suitable (presenting) else tall graph
        """
        l = ['digraph { ',
             # 'rankdir=LR',  # XX I really wish to have TB, with rows: IP, plaintext, timestamp
             'label="Convey field types (dashed nodes = identifiable automatically, dashed edges = field identity)"']
        used = set()
        formatting_disabled = "[color=lightgray fontcolor=lightgray]" if flags & 1 else ""
        loop = []
        existing_edges = set()

        def add(start, target, disable_s):
            if (start, target) not in used:
                used.add((start, target))
                used.add((target, start))
                if not disable_s:
                    used.update([start, target])
                l.append(f"{start} -> {target}{disable_s}")
                if (target, start) in existing_edges:
                    l[-1] += "[dir=both]"
                return True
            return False

        for enabled, ((start, target), m) in [(True, f) for f in methods.items()] + [(False, f) for f in
                                                                                     methods_deleted.items()]:
            loop.append((enabled, start, target, m))
            existing_edges.add((start, target))
        # for enabled, ((start, target), m) in [(True, f) for f in methods.items()] + [(False, f) for f in methods_deleted.items()]:
        for enabled, start, target, m in loop:
            disable_s = "" if enabled else formatting_disabled
            if start.group != target.group and target.group.name != target and target.group in [TypeGroup.dns,
                                                                                                TypeGroup.nmap,
                                                                                                TypeGroup.custom]:
                # Every type that goes to ex: `dns`, continues to all `dns` subtypes. We want `hostname -> spf` to go through `dns`.
                # This is not the case of the group `whois` - this group has explicitly stated the path in methods,
                #   ex: `ip → whois → country` but also `phone → country`.
                add(start, target.group.name, disable_s)
                start = target.group.name
            if add(start, target, disable_s):
                if m is True:
                    l[-1] += "[style=dashed]"
                if isinstance(m, PickMethod):
                    l[-1] += '[label="choose ' + ",".join([name for name, _ in m.get_options()]) + '" fontsize=10]'
                elif isinstance(m, PickInput):
                    l[-1] += f'[label="{m.parameter_name}" fontsize=10]'
        for tg in TypeGroup:
            if tg is not TypeGroup.general:
                if tg.name in used:
                    l.append(f"{tg.name} [shape=box]")
                else:
                    l.append(f"{tg.name} [shape=box]{formatting_disabled}")
        for f in types:
            disable_s = "" if f in used else formatting_disabled
            label = [f.name]
            s = ""
            if f.description:
                label.append(f.description)
            if flags & 2:
                if f.usual_names:
                    label.append("usual names: " + ", ".join(f.usual_names))
                s = ' [label="' + r"\n".join(label) + '"]'
            l.append(f'{f.name}{s}{disable_s}')
            if f in Types.get_guessable_types():
                l.append(f'{f.name} [style=dashed]')
            # if f.is_private:
            #    l.append(f'{f.name} [shape=box]')

        # formatting mark to put one cluster below another, not aside
        if flags & 16:  # looks nicer in a presentation
            l.extend([
                # "spf -> timestamp[style=invis]",
                # "formatted_time -> plaintext[style=invis]",
            ])
        else:  # looks nicer in README.md
            l.extend([
                "port -> urlencode[style=invis]",
                "charset -> timestamp[style=invis]"
            ])
        l.append("}")
        return "\n".join(l)

    @staticmethod
    def _get_methods():
        """  These are known methods to compute a field from another field.
            They should return scalar or list.

            Method ~ lambda: Will be processed when converting one type to another
            Method ~ None: Will be skipped. (TypeGroup.custom fields usually have None.)
            Method ~ True: The user should not be offered to compute from the field to another from a longer distance.
                However if they already got the first type, it is the same as if they had the other.
                Example: (source_ip → ip) hostname from source_ip will be computed as if from an ip,
                Example: (abusemail → email) We do not want to offer the conversion hostname → whois → abusemail → email → hostname,
                    in other words we do not offer to compute a hostname from a hostname, however when selecting
                    an existing abusemail column, we offer conversion abusemail → (invisible email) → hostname
        """

        t = Types
        return {
            (t.source_ip, t.ip): True,
            (t.destination_ip, t.ip): True,
            (t.any_ip, t.ip): any_ip_ip,  # "91.222.204.175 93.171.205.34" -> "91.222.204.175" OR '"1.2.3.4"' -> 1.2.3.4
            (t.port_ip, t.ip): port_ip_ip,  # IP written with a port 91.222.204.175.23 -> 91.222.204.175
            (t.port_ip, t.port): port_ip_port,
            (t.url, t.hostname): url_hostname,
            (t.url, t.port): url_port,
            (t.hostname, t.ip): Checker.hostname_ips if Config.get("multiple_hostname_ip",
                                                                   "FIELDS") else Checker.hostname_ip,
            # (t.url, t.ip): Whois.url2ip,
            (t.ip, t.whois): Whois,
            # (t.asn, t.whois): Whois, # XX can be easily allowed, however Whois object will huff there is no IP prefix range
            (t.cidr, t.ip): Checker.cidr_ips if Config.get("multiple_cidr_ip", "FIELDS") else
            lambda x: str(ipaddress.ip_interface(x).ip),
            (t.whois, t.prefix): lambda x: str(x.get[0]),
            (t.whois, t.asn): lambda x: x.get[3],
            (t.whois, t.abusemail): lambda x: x.get[6],
            (t.whois, t.country): lambda x: x.get[5],
            (t.whois, t.netname): lambda x: x.get[4],
            (t.whois, t.csirt_contact):
                lambda x: Contacts.country2mail[x.get[5]] if x.get[5] in Contacts.country2mail else "",
            (t.whois, t.incident_contact): lambda x: x.get[2],
            (t.abusemail, t.cc_contact): lambda x: "; ".join(
                Contacts.mail2cc[domain] for domain in Contacts.get_domains(x) if domain in Contacts.mail2cc
            ),
            (t.plaintext, t.bytes): lambda x: x.encode("UTF-8"),
            (t.bytes, t.plaintext): Checker.bytes_plaintext,
            (t.bytes, t.base64): lambda x: b64encode(x).decode("UTF-8"),
            (t.base64, t.bytes): b64decode,
            (t.bytes, t.quoted_printable): lambda x: encodestring(x).decode("UTF-8"),
            (t.quoted_printable, t.bytes): decodestring,
            (t.urlencode, t.plaintext): lambda x: unquote(x),
            (t.plaintext, t.urlencode): lambda x: quote(x),
            (t.plaintext, t.external): None,
            (t.plaintext, t.code): None,
            (t.plaintext, t.reg): None,
            (t.bytes, t.charset): Checker.decode,
            (t.reg, t.reg_s): None,
            (t.reg, t.reg_m): None,
            (t.wrong_url, t.url): wrong_url_2_url,
            (t.hostname, t.url): lambda x: "http://" + x,
            # X (t.ip, t.url): lambda x: "http://" + x,
            (t.ip, t.hostname): lambda x: socket.gethostbyaddr(x)[0],
            (t.url, t.web): Web,
            (t.web, t.http_status): lambda x: x.get[0],
            (t.web, t.text): lambda x: x.get[1],
            (t.web, t.html): lambda x: x.get[2],
            (t.web, t.redirects): lambda x: x.get[3],
            (t.web, t.x_frame_options): lambda x: x.get[4],
            (t.web, t.csp): lambda x: x.get[5],
            (t.web, t.form_names): lambda x: x.get[6],
            (t.hostname, t.ports): nmap,
            (t.ip, t.ports): nmap,
            (t.ports, t.port): True,
            (t.hostname, t.spf): dig("SPF"),
            (t.hostname, t.txt): dig("TXT"),
            (t.hostname, t.a): dig("A"),
            (t.hostname, t.aaaa): dig("AAAA"),
            (t.hostname, t.ns): dig("NS"),
            (t.hostname, t.mx): dig("MX"),
            (t.hostname, t.dmarc): dig("DMARC"),
            (t.abusemail, t.email): True,
            (t.email, t.hostname): lambda x: x[x.index("@") + 1:],
            # (t.email, check legit mailbox)
            (t.country_name, t.country): address_country,
            (t.country, t.country_name): lambda x: country_codes[x],
            (t.country, t.csirt_contact): lambda x: Contacts.country2mail[x] if x in Contacts.country2mail else "",
            (t.phone, t.country): phone_country,
            (t.hostname, t.tld): Checker.HostnameTld,
            (t.prefix, t.cidr): Checker.prefix_cidr,
            (t.redirects, t.url): True,
            (t.tld, t.country): True,
            (t.timestamp, t.formatted_time): Checker.time_format,
            (t.timestamp, t.isotimestamp): Checker.isotimestamp,
            (t.timestamp, t.date): Checker.date,
            (t.timestamp, t.time): lambda val: Checker.parse_timestamp(val).time(),
            (t.unit, t.plaintext): Checker.unit_expand,
            (t.plaintext, t.unit): Checker.unit_expand
        }


def get_module_from_path(path):
    if not Path(path).is_file():
        return False
    spec = importlib.util.spec_from_file_location("", path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def url_ip(url):  # not used right now
    """ Shorten URL to domain, find out related IPs list. """
    url = urlparse(url.strip())  # "foo.cz/bar" -> "foo.cz", "http://foo.cz/bar" -> "foo.cz"
    uri = url.hostname if url.scheme else url.path
    try:
        return socket.gethostbyname(uri)  # returns 1 address only, we do not want all of them
    except socket.gaierror as e:
        # logger.warning("Socket gethostbyname error for URI {} .".format(uri))
        # Config.error_caught()
        return None
    # if we wanted all of the IPs:
    # recs = socket.getaddrinfo(uri, 0, 0, 0, socket.IPPROTO_TCP)
    # result = []
    # for ip in recs:
    #    result.append(ip[4][0])
    # return result


def url_hostname(url):
    """ Covers both use cases "http://example.com..." and "example.com..." """
    s = urlsplit(url)
    s = s.netloc or s.path.split("/")[:1][0]
    return s.split(":")[0]  # strips port



# The file is not sourced from __init__ like decorators.py (where we hunt every ms)
# so there is no problem if put here many functions that would cause a bottleneck
import logging
import sys
from threading import Thread, Event, current_thread
from time import sleep
from typing import Callable

from .config import Config

logger = logging.getLogger(__name__)


def print_atomic(s):
    """ print not atomic, \n trailed when threading
        If in thread, prints its name.
        If quiet, no output is given.
    """
    if Config.is_quiet():
        return
    n = current_thread().name
    t = "" if n == "MainThread" else "/" + n + "/ "
    sys.stdout.write(t + s + "\n")


def timeout(seconds: int, function: Callable, *args, **kwargs):
    """
    Launch the function in a new thread. If function stops before timeout, returns output or re-raise an exception received.
    If timeout reached, raises TimeoutError and let the thread hang indefinitely, ignoring any successive output or exception.
    However, thread can print to stdout even after TimeoutError has been reached because there is no safe option to kill a thread.
    :param seconds: int
    :param function:
    :param args: Any positional arguments the function receives.
    :param kwargs: Any keyword arguments the function receives.
    :return:
    """
    result = []
    exception = []

    def wrapper(*args, **kwargs):
        try:
            result.append(function(*args, **kwargs))
        except Exception as e:
            exception.append(e)

    thread = Thread(target=wrapper, args=args, kwargs=kwargs)
    thread.daemon = True  # will not block the program exit if hung
    thread.start()
    thread.join(seconds)
    if thread.is_alive():
        raise TimeoutError(f'Timeout {seconds} of {function}')
    else:
        if exception:
            raise exception[0] from None
        return result[0]


def _lazy_print(timeout, msg, done):
    sleep(timeout)
    if not done.is_set():
        logger.info(msg)


def lazy_print(msg: str, timeout: float = 1) -> Event:
    """
    Print message only if time out reached and Event not marked with .set().
    Example:
        l = lazy_print(1, "... still loading")
        l.set() # will not print anything if called under 1 sec
    """
    event = Event()
    Thread(target=_lazy_print, args=(timeout, msg, event)).start()
    return event



#!/usr/bin/env python3
import bdb
import os
import socket
import sys
from sys import exit

from .ipc import send, recv, socket_file, daemon_pid

__doc__ = """Convey – CSV swiss knife brought by CSIRT.cz"""
__author__ = "Edvard Rejthar, CSIRT.CZ"
__date__ = "$Feb 26, 2015 8:13:25 PM$"

if sys.version_info[0:2] < (3, 7):
    print("We need at least Python 3.7, your version is " + sys.version + ". Try an older Convey release or rather upgrade Python.")
    exit()


def main():
    if not sys.stdin.isatty():
        # this is not a terminal - we may already nr receiving something through a pipe
        # load it immediately and not in the wrapper because daemon could suck the pipe of the instance
        # however sys.stdin.read() is a blocking operation so that we stuck in a pipe without stdin
        #    ex: running tests from the IDE or launching subprocess.run("convey ...")
        # As a compromise we check current sys.argv if there is an input token or not.
        # Note that this way we still get stuck when
        # launching with an implicit input `subprocess.run("convey file.csv")`,
        # we have to explicitly state `subprocess.run("convey --file file.csv")`
        if not any(x in sys.argv for x in ['-i', '--input', '--file']):
            sys.argv.extend(["--input", sys.stdin.read().rstrip()])
    daemonize_on_exit = True
    if os.path.exists(socket_file):  # faster than importing Pathlib.path
        try:
            pipe = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            pipe.connect(socket_file)
            pipe.settimeout(10)  # it can take several seconds to call whois
        except ConnectionRefusedError:
            pass
        else:
            send(pipe, repr([os.getcwd()] + sys.argv))
            try:
                response = recv(pipe)
            except socket.timeout:
                print("It seems daemon is stuck. You may kill it with pkill `convey`.")
                from .config import Config
                if Config.get("github_crash_submit"):
                    Config.github_issue(f"daemon stuck", "Command line:\n```bash\n" + repr(sys.argv) + "\n```")
            else:
                # chr(4) at the end means this was not a single query check and we should load full convey libraries
                if type(response) is str:
                    if response.endswith(chr(3)):  # daemon has missing input
                        pass
                    elif not response.endswith(chr(17)):  # daemon is not stopping
                        daemonize_on_exit = False
                        if not response.endswith(chr(4)):  # daemon brings some results
                            print(response, end='')
                            exit()
            finally:
                pipe.close()

    try:
        from .controller import Controller
        Controller().run()
    except KeyboardInterrupt:
        print("Interrupted")
    except SystemExit as e:
        if daemonize_on_exit:
            from .config import Config
            try:
                if Config.get("daemonize", get=bool) and not daemon_pid():
                    # we are sure there is no running instance of the daemon
                    import subprocess
                    subprocess.Popen([sys.argv[0], '--daemon', 'server'], stdout=subprocess.DEVNULL)
            except FileNotFoundError:
                # XX we should check this code allows Win to run without daemon or rather make the daemon work on Win
                pass
    except bdb.BdbQuit:
        pass
    except:
        import traceback
        debug = True
        try:
            from .config import Config
            debug = Config.is_debug() or Config.get("crash_post_mortem")
        except ImportError:
            Config = None

        type_, value, tb = sys.exc_info()
        if debug:
            traceback.print_exc()
            import pdb
            mod = Config.get_debugger() if Config else pdb
            mod.post_mortem(tb)
        elif Config:
            # XX we should check the -3 line has something to do with the convey directory, otherwise print out other lines
            #   (we are not interested of the errors in other libraries)
            print(f"Convey crashed at {value} on {traceback.format_exc().splitlines()[-3].strip()}")
            if Config.get("github_crash_submit"):
                body = f"```bash\n{traceback.format_exc()}```\n\n```json5\n{tb.tb_next.tb_frame.f_locals}\n```"
                Config.github_issue(f"crash: {value}", body)


if __name__ == "__main__":
    main()
